/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataCacheImplConfiguration;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.config.DtoImplConfiguration;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import org.apache.commons.lang3.RandomStringUtils;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import javax.sql.DataSource;
import java.time.LocalDateTime;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Predicate;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

@CbsTest
@SpringJUnitConfig({DataCacheImplConfiguration.class, ReaderWriterEndToEndIT.DBConfig.class, DtoImplConfiguration.class})
@Disabled("Requires running mariadb with liquibase changes applied")
class ReaderWriterEndToEndIT {

    @Autowired
    private Writer<VaccinationRecord<Vaccination>> vaccinationWriter;

    @Autowired
    private Reader<Vaccination, String> vaccinationReader;


    @Test
    void check_constraint() {
        final VaccinationRecordImpl<Vaccination> record = record(randomCpr(), null, null);
        assertThrows(RuntimeException.class, () -> vaccinationWriter.write(record));
    }

    @Test
    void data_record_is_returned() {
        final String cpr = randomCpr();

        write(record(cpr, null, 15L));

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    private String randomCpr() {
        final String cpr = RandomStringUtils.randomNumeric(10);
        System.out.println("cpr = " + cpr);
        return cpr;
    }

    @Test
    void data_record_with_time_to_live_in_future_is_returned() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now().plusMinutes(60L), 15L));

        assertRecordReturnedWithDrugId(cpr, 15L);

    }

    @Test
    void data_record_with_time_to_live_in_past_is_not_returned() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now(), 15L));

        assertNoRecordReturned(cpr);
    }

    @Test
    void invalidation_record_is_not_returned() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now(), null));

        assertNoRecordReturned(cpr);
    }

    @Test
    void future_invalidation_record_is_not_returned() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), null));

        assertNoRecordReturned(cpr);
    }

    @Test
    void data_is_invalidated() {
        final String cpr = randomCpr();

        write(record(cpr, null, 15L));
        write(record(cpr, LocalDateTime.now(), null));

        assertNoRecordReturned(cpr);
    }

    @Test
    void invalidated_data_is_replaced() {
        final String cpr = randomCpr();

        write(record(cpr, null, 1L));
        write(record(cpr, LocalDateTime.now(), null));
        write(record(cpr, null, 15L));

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    @Test
    void data_is_still_valid_after_future_invalidation() {
        final String cpr = randomCpr();

        write(record(cpr, null, 15L));
        write(record(cpr, LocalDateTime.now().plusMinutes(10L), null));

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    @Test
    void new_invalidation_wins_over_previous_time_to_live() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L));

        write(record(cpr, LocalDateTime.now(), null));

        assertNoRecordReturned(cpr);
    }

    @Test
    void check_sequences_0() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now(), null));

        write(record(cpr, null, 1L));

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L)); // <-- this wins

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    @Test
    void check_sequences_1() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now(), null));

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L));

        write(record(cpr, null, 15L));  // <-- this wins

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    @Test
    void check_sequences_2() {
        final String cpr = randomCpr();

        write(record(cpr, null, 1L));

        write(record(cpr, LocalDateTime.now(), null));

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L)); // <-- this wins

        assertRecordReturnedWithDrugId(cpr, 15L);
    }

    @Test
    void check_sequences_3() {
        final String cpr = randomCpr();

        write(record(cpr, null, 15L));

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L));

        write(record(cpr, LocalDateTime.now(), null)); // <-- this wins

        assertNoRecordReturned(cpr);
    }

    @Test
    void check_sequences_4() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 15L));

        write(record(cpr, null, 15L));

        write(record(cpr, LocalDateTime.now(), null)); // <-- this wins

        assertNoRecordReturned(cpr);
    }

    @Test
    void check_sequences_5() {
        final String cpr = randomCpr();

        write(record(cpr, LocalDateTime.now().plusMinutes(10L), 1L));

        write(record(cpr, LocalDateTime.now(), null));

        write(record(cpr, null, 15L)); // <-- this wins

        assertRecordReturnedWithDrugId(cpr, 15L);
    }


    private void assertRecordReturnedWithDrugId(String cpr, Long drugId) {

        sleep();
        final Vaccination vaccination = vaccinationReader.search(cpr);
        assertNotNull(vaccination);
        assertTrue(vaccination.vaccinations().stream().map(EffectuatedVaccination::drugIdentifier).anyMatch(Predicate.isEqual(drugId)));
    }

    private void sleep() {
        try {
            Thread.sleep(1000L);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            e.printStackTrace();
        }
    }


    private void assertNoRecordReturned(String cpr) {
        sleep();
        assertNull(vaccinationReader.search(cpr));
    }

    private void write(VaccinationRecordImpl<Vaccination> record) {
        sleep();
        assertTrue(vaccinationWriter.write(record));
    }

    private VaccinationRecordImpl<Vaccination> record(String cpr, LocalDateTime invalidationTime, Long drugIdentifier) {
        return new VaccinationRecordImpl<Vaccination>()
            .cpr(cpr)
            .traceId(UUID.randomUUID().toString())
            .invalidationTime(invalidationTime)
            .setData(Optional.ofNullable(drugIdentifier)
                .map(drugId -> VaccinationBuilder.vaccination(cpr, false)
                    .add(EffectuatedVaccinationBuilder.effectuatedVaccination(drugId, LocalDateTime.now()).build())
                    .build())
                .orElse(null));
    }

    @Configuration
    public static class DBConfig {
        @Bean
        public DataSource dataSource() {
            DriverManagerDataSource dataSource = new DriverManagerDataSource();
            dataSource.setUrl("jdbc:mysql://localhost:3306/cbs-datacache");
            dataSource.setUsername("admin");
            dataSource.setPassword("admin");
            return dataSource;
        }
    }
}
