/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluations;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;

import java.time.LocalDateTime;
import java.util.Comparator;
import java.util.Objects;
import java.util.stream.Stream;

public class PassportEvaluationsImpl implements PassportEvaluations {
    private final PassportEvaluation vaccinationEvaluation;
    private final PassportEvaluation testPassportEvaluation;
    private final RecoveryPassportEvaluation recoveryPassportEvaluation;
    private final LocalDateTime expiresAfter;
    private final boolean passportGranted;
    private final String rejectionReason;

    public PassportEvaluationsImpl(PassportEvaluation vaccinationEvaluation, PassportEvaluation testPassportEvaluation, RecoveryPassportEvaluation recoveryPassportEvaluation) {

        this.vaccinationEvaluation = vaccinationEvaluation;
        this.testPassportEvaluation = testPassportEvaluation;
        this.recoveryPassportEvaluation = recoveryPassportEvaluation;

        expiresAfter = Stream.of(vaccinationEvaluation, testPassportEvaluation, recoveryPassportEvaluation)
            .map(PassportEvaluation::expiresAfter)
            .filter(Objects::nonNull)
            .max(Comparator.naturalOrder()).orElse(null);
        passportGranted = Stream.of(vaccinationEvaluation, testPassportEvaluation, recoveryPassportEvaluation)
            .anyMatch(PassportEvaluation::passportGranted);
        rejectionReason = Stream.of(vaccinationEvaluation, testPassportEvaluation, recoveryPassportEvaluation)
            .map(PassportEvaluation::rejectionReason).filter(Objects::nonNull)
            .reduce("",
                (a, r) -> "".equals(a) ? r : a + "," + r,
                (x, y) -> x + "," + y);
    }

    @Override
    public boolean passportGranted() {
        return passportGranted;
    }

    @Override
    public LocalDateTime expiresAfter() {
        return expiresAfter;
    }

    @Override
    public String rejectionReason() {
        return rejectionReason;
    }

    @Override
    public PassportEvaluation vaccinationPassport() {
        return vaccinationEvaluation;
    }

    @Override
    public PassportEvaluation testPassport() {
        return testPassportEvaluation;
    }

    @Override
    public RecoveryPassportEvaluation recoveryPassport() {
        return recoveryPassportEvaluation;
    }
}
