/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForVaccinationPass;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.afterMax;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.afterMin;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.beforeMax;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.beforeMin;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.expireDenied;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.expireGranted;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.on;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positiveAntigen;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positiveAntigenBeforeNegativePcr;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positivePcr;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.completedMultiDose;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.partial;
import static java.util.stream.Stream.of;
import static org.junit.jupiter.api.Assertions.assertEquals;

class VaccinationPassportEvaluatorTest {

    @SuppressWarnings("unused")
    @DisplayName("evaluation")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateSource")
    void evaluate(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final VaccinationPassportConfiguration configuration,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final VaccinationPassportEvaluator evaluator = new VaccinationPassportEvaluator(configuration);
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        System.out.println(evaluation.rejectionReason());
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateSource() {
        return of(
            defaultForVaccinationPass("No vaccinations").denied().build(),
            defaultForVaccinationPass("Completed vaccination where less than minimum period has passed").vaccination(completedMultiDose(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where less than maximum period has passed").vaccination(completedMultiDose(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly maximum period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where more than maximum period has passed").vaccination(completedMultiDose(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where more than suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is now").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW)).denied().expire(expireDenied(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is in future").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW.plusNanos(1))).denied().expire(expireDenied(DEFAULT_NOW.plusNanos(1), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where more than suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed and negative pcr test after positive antigen").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigenBeforeNegativePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with one vaccination where exactly minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly minimum period has passed and 2nd vaccination where less than minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly maximum minus minumum period has passed and 2nd vaccination administered now").vaccination(partial(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_NOW)).granted().expire(expireGranted(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build()

        );
    }
}
