/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.cache;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.MissingInformationException;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

public class DataReader {

    private final NamedParameterJdbcTemplate jdbcTemplate;
    private final ObjectMapper objectMapper;
    private final boolean includeChildren;

    public DataReader(NamedParameterJdbcTemplate jdbcTemplate, ObjectMapper objectMapper, boolean includeChildren) {
        this.jdbcTemplate = jdbcTemplate;
        this.objectMapper = objectMapper;
        this.includeChildren = includeChildren;
    }

    public Collection<Data> getData(String cpr) {
        final List<Data> data = jdbcTemplate.query(
            "select p.person_identifier, p.givenName, p.familyName , v.data , l.data " +
                "from stamdata_person p " +
                "left join vaccination_view v on v.cpr = p.person_identifier " +
                "left join lab_result_view l on l.cpr = p.person_identifier " +
                "where p.person_identifier = :cpr" +
                (includeChildren ? " or p.person_identifier in (select c.child_identifier from stamdata_child_relation c where c.person_identifier = :cpr)" : ""),
            new MapSqlParameterSource("cpr", cpr),
            (rs, rowNum) -> new Data(
                map(rs.getBytes("v.data"), Vaccination.class),
                map(rs.getBytes("l.data"), LabResults.class),
                new PersonRecordImpl()
                    .personIdentifier(rs.getString("p.person_identifier"))
                    .givenName(rs.getString("p.givenName"))
                    .familyName(rs.getString("p.familyName")))
        );

        if (data.stream().noneMatch(d -> Objects.equals(d.personRecord().personIdentifier(), cpr))) {
            throw new MissingInformationException("Person data is missing");
        }

        final Function<Data, Custody> custodyFunction = custodyFunction(cpr);

        data.forEach(d -> d.custody(custodyFunction.apply(d)));

        return data;
    }

    private Function<Data, Custody> custodyFunction(String parentCpr) {
        AtomicInteger counter = new AtomicInteger(0);

        if (!includeChildren) {
            return data -> null;
        }

        return data -> parentCpr.equals(data.personRecord().personIdentifier())
            ? Custody.parent()
            : Custody.child(counter.incrementAndGet());
    }


    private <T> T map(byte[] data, Class<T> valueType) {
        if (data == null || data.length == 0) {
            return null;
        }

        try {
            return objectMapper.readValue(data, valueType);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }


}
