/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.handler;

import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CertificateSigningException;
import dk.sundhedsdatastyrelsen.cbs.service.model.SignedCertificateCWT;
import dk.sundhedsdatastyrelsen.cbs.service.passports.ChecksumUtils;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestSample;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestType;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import se.digg.dgc.payload.v1.DGCSchemaException;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.service.DGCEncoder;

import java.io.IOException;
import java.security.SignatureException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.temporal.TemporalAmount;
import java.util.Collections;
import java.util.UUID;

@Component
public class EUCovidTestCertificateSigningHandler {

    private final DGCEncoder encoder;
    private final EUAntigenFiltration euAntigenFiltration;
    private final TemporalAmount maxValidity;

    public EUCovidTestCertificateSigningHandler(
        DGCEncoder encoder,
        EUAntigenFiltration euAntigenFiltration,
        @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${eu.covid.test.certificate.signing.maximum-validity:PT72H}')}")
            TemporalAmount maxValidity
    ) {
        this.encoder = encoder;
        this.euAntigenFiltration = euAntigenFiltration;
        this.maxValidity = maxValidity;
    }

    public SignedCertificateCWT getSignedCertificateCWT(GetSignedEUCovidTestCertificateRequest request, AuditData auditData) {
        auditData
            .withSensitivePersonalInformation("personIdentifier", request.getPersonIdentifier());

        final TestSample payload = request.getPayload();

        if (payload.getTypeOfTest().equals(TestType.RAT) && isNotApproved(payload)) {
            throw new CertificateSigningException(ErrorMessages.NOT_HSC_APPROVED);
        }

        final Instant sampleCollected = payload.getSampleCollected().toInstant();
        final Instant expiration = sampleCollected.plus(maxValidity);

        if (expiration.isBefore(Instant.now())) {
            throw new CertificateSigningException(ErrorMessages.TEST_IS_EXPIRED);
        }

        final DigitalGreenCertificate certificate = PassportUtils.buildDigitalGreenCertificate(payload.getGivenName(), payload.getFamilyName(), payload.getDateOfBirth());

        final String ci = ChecksumUtils.generateUVCI(UUID.randomUUID().toString());

        certificate.setT(Collections.singletonList(
            new TestEntry()
                .withIs(VaccineMapper.ISSUER)
                .withCo(payload.getCountryOfTest().value())
                .withCi(ci)
                .withTg(payload.getDiseaseAgentTargeted().value())
                .withSc(sampleCollected)
                .withMa(payload.getManufacturer())
                .withTt(payload.getTypeOfTest().value())
                .withTc(payload.getTestingCenter())
                .withTr(payload.getTestResult().value())
        ));

        final String cwt = encode(certificate, expiration);

        auditData
            .withNonPersonalInformation("CI", ci)
            .withSensitivePersonalInformation("CWT", cwt);

        return new SignedCertificateCWT(cwt, ci);
    }

    private boolean isNotApproved(TestSample payload) {
        return !euAntigenFiltration.isIdDeviceLegalAtTime(
            payload.getManufacturer(),
            payload.getSampleCollected().withZoneSameInstant(ZoneId.systemDefault()).toLocalDateTime()
        );
    }

    private String encode(DigitalGreenCertificate certificate, Instant expiration) {
        try {
            return encoder.encode(certificate, expiration);
        } catch (DGCSchemaException | IOException | SignatureException e) {
            throw new CertificateSigningException(ErrorMessages.ENCODING_ERROR, e);
        }
    }

}
