/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.RecoveryPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryLabResult;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CwtException;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DkRules;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EnhancedTestEntry;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EnhancedVaccinationEntry;
import dk.sundhedsdatastyrelsen.cbs.service.util.AuditUtils;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import org.springframework.lang.NonNull;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.payload.v1.VaccinationEntry;
import se.digg.dgc.service.DGCEncoder;

import java.time.*;
import java.time.temporal.TemporalAmount;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody.getLabel;

public class DkInfoPassportFactory implements PassportFactory {
    public static final String DKINFO_HEADER = "DK3:";
    private final VaccineMapper vaccineMapper;
    private final DGCEncoder encoder;
    private final RecoveryPassportEvaluator dkRecoveryPassportEvaluator;
    private final TemporalAmount vaccinationValidityFromEffectuationTime;
    private final TemporalAmount partialVaccinationValidityFromEffectuationTime;
    private final TemporalAmount pcrTestValidityFromSamplingTime;
    private final TemporalAmount agTestValidityFromSamplingTime;
    private final TemporalAmount labResultMaxAge;

    public DkInfoPassportFactory(VaccineMapper vaccineMapper, DGCEncoder encoder, @DkRules RecoveryPassportEvaluator dkRecoveryPassportEvaluator, TemporalAmount vaccinationValidityFromEffectuationTime, TemporalAmount partialVaccinationValidityFromEffectuationTime, TemporalAmount pcrTestValidityFromSamplingTime, TemporalAmount agTestValidityFromSamplingTime, TemporalAmount labResultMaxAge) {
        this.vaccineMapper = vaccineMapper;
        this.encoder = encoder;
        this.dkRecoveryPassportEvaluator = dkRecoveryPassportEvaluator;
        this.vaccinationValidityFromEffectuationTime = vaccinationValidityFromEffectuationTime;
        this.partialVaccinationValidityFromEffectuationTime = partialVaccinationValidityFromEffectuationTime;
        this.pcrTestValidityFromSamplingTime = pcrTestValidityFromSamplingTime;
        this.agTestValidityFromSamplingTime = agTestValidityFromSamplingTime;
        this.labResultMaxAge = labResultMaxAge;
    }

    @Override
    @NonNull
    public Set<CoronaPass> createPassportFrom(@NonNull Vaccination vaccination, @NonNull LabResults labResults, PersonRecord person, AuditData auditData, Custody custody) {
        final DigitalGreenCertificate certificate = PassportUtils.buildCertificateFrom(person);

        final List<TestEntry> testEntries = getTestEntries(labResults);
        if (!testEntries.isEmpty()) {
            certificate.setT(testEntries);
        }

        final List<VaccinationEntry> vaccinationEntries = getVaccinationEntries(vaccination);
        if (!vaccinationEntries.isEmpty()) {
            certificate.setV(vaccinationEntries);
        }

        //todo refactor and somehow use the RecoveryPassportEvaluation object received in DkPassportFactory instead of re-running evaluation here
        final List<RecoveryEntry> recoveryEntries = getRecoveryEntries(dkRecoveryPassportEvaluator.evaluate(vaccination, labResults, LocalDateTime.now()));
        if (!recoveryEntries.isEmpty()) {
            certificate.setR(recoveryEntries);
        }

        final CoronaPass coronaPass = new CoronaPass(PassType.DKINFO, encode(certificate));
        if (custody != null) {
            coronaPass.setProperty("Custody", custody.getLabel());
        }

        final Set<CoronaPass> coronaPasses = Collections.singleton(coronaPass);
        AuditUtils.addDkInfoPassportAuditData(coronaPasses, auditData, getLabel(custody));
        return coronaPasses;
    }

    @NonNull
    private List<VaccinationEntry> getVaccinationEntries(Vaccination vaccination) {
        return vaccineMapper.toVaccinationEntries(vaccination)
            .stream()
            .map(this::withValidity)
            .collect(Collectors.toList());
    }

    private EnhancedVaccinationEntry withValidity(VaccinationEntry entry) {
        final LocalDateTime effectuationTime = entry.getDt().atStartOfDay();
        final ZoneOffset offset = ZoneId.systemDefault().getRules().getOffset(effectuationTime);
        final Instant validTo = effectuationTime
            .plus(Objects.equals(entry.getDn(), entry.getSd()) ? vaccinationValidityFromEffectuationTime : partialVaccinationValidityFromEffectuationTime)
            .toInstant(offset);
        return new EnhancedVaccinationEntry(entry, validTo);
    }

    @NonNull
    private List<TestEntry> getTestEntries(LabResults labResults) {
        final LocalDateTime now = LocalDateTime.now();
        return labResults.results().stream()
            .filter(LabResult::negative)
            .filter(result -> result.samplingDateTime().isAfter(now.minus(labResultMaxAge)))
            .map(this::toTestEntry)
            .collect(Collectors.toList());
    }

    @NonNull
    private List<RecoveryEntry> getRecoveryEntries(RecoveryPassportEvaluation recoveryEvaluation) {
        final RecoveryLabResult recoveryLabResult = recoveryEvaluation.result();
        if (!recoveryEvaluation.passportGranted() || recoveryLabResult == null) {
            return Collections.emptyList();
        }
        final RecoveryEntry recoveryEntry = RecoveryMapper.toRecovery(
            recoveryLabResult.samplingDateTime().toLocalDate(),
            recoveryLabResult.recoveryValidFrom().toLocalDate(),
            recoveryLabResult.recoveryValidUntil().toLocalDate()
        );
        return Collections.singletonList(recoveryEntry);
    }

    private EnhancedTestEntry toTestEntry(LabResult result) {
        final TestEntry entry = LabResultMapper.toTestEntry(result);
        final TemporalAmount validityFromSamplingTime;
        if (result instanceof LabResult.PcrResult) {
            validityFromSamplingTime = pcrTestValidityFromSamplingTime;
        } else if (result instanceof LabResult.AntigenResult) {
            validityFromSamplingTime = agTestValidityFromSamplingTime;
        } else {
            throw new UnsupportedOperationException("Unknow test type: " + result.type());
        }
        final ZoneOffset offset = ZoneId.systemDefault().getRules().getOffset(result.samplingDateTime());
        return new EnhancedTestEntry(entry, result.samplingDateTime().plus(validityFromSamplingTime).toInstant(offset));
    }

    private String encode(DigitalGreenCertificate certificate) {
        try {
            return encoder.encode(certificate, Instant.now().plus(Duration.ofHours(72L))).replaceFirst(DGCConstants.DGC_V1_HEADER, DKINFO_HEADER);
        } catch (Exception e) {
            throw new CwtException("CWT Encoding Error", e);
        }
    }
}
