/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.MoreCollectors;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaProducer;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditLogger;
import dk.sundhedsdatastyrelsen.cbs.service.config.*;
import dk.sundhedsdatastyrelsen.cbs.service.cache.Data;
import dk.sundhedsdatastyrelsen.cbs.service.cache.DataReader;
import dk.sundhedsdatastyrelsen.cbs.service.config.*;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPassResult;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.CombinedPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.DkPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.EuPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.MockDataService;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.jdbc.datasource.lookup.JndiDataSourceLookup;
import org.springframework.lang.NonNull;
import org.springframework.mock.jndi.SimpleNamingContextBuilder;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.service.impl.MrzParser;
import se.digg.dgc.signatures.cwt.Cwt;

import javax.naming.NamingException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalAmount;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
@SpringJUnitConfig({EUAntigenSpecConfiguration.class, CoronaPassServiceTest.MockConfig.class, CombinedPassportFactory.class, DkPassportFactory.class, EuPassportFactory.class, SignatureConfiguration.class, DkPassportConfiguration.class, EuPassportConfiguration.class, VaccineMapperConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@TestPropertySource(properties = {
    "cwt.eu.ag-result.maximum-age=P1W",
    "cwt.eu.pcr-result.maximum-age=P2W",
    "cwt.dk.maximum-validity=PT24H",
    "cwt.eu.maximum-validity=PT48H",
    "eu.recovery.validator.minimum-period-after-recovery=P11D",
    "eu.recovery.validator.maximum-period-after-recovery=P180D"
})
class CoronaPassServiceTest {
    private static final long DRUG_ID = 32138956351L;
    private static final RandomData randomData = new RandomData(System.currentTimeMillis());
    private final String idwsMessageId = UUID.randomUUID().toString();
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.dk.maximum-validity}')}")
    private TemporalAmount dkMaximumValidity;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.eu.maximum-validity}')}")
    private TemporalAmount euMaximumValidity;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.eu.pcr-result.maximum-age}')}")
    private TemporalAmount euPcrMaxAge;
    @Mock
    private CbsKafkaProducer cbsKafkaProducer;
    @Autowired
    private CombinedPassportFactory passportFactory;
    @Autowired
    private AuditLogger auditLogger;

    @Mock
    private DataReader dataReader;

    private CoronaPassService coronaPassService;

    @Configuration
    public static class MockConfig {
        @Bean
        @Primary
        AuditLogger auditLogger() {
            return Mockito.mock(AuditLogger.class);
        }

    }

    @BeforeEach
    void setUp() {
        coronaPassService = new CoronaPassService(cbsKafkaProducer, passportFactory, auditLogger, dataReader);
    }

    @BeforeAll
    static void setUpJNDI() throws NamingException {
        //noinspection deprecation
        SimpleNamingContextBuilder.emptyActivatedContextBuilder();
        new JndiDataSourceLookup().getJndiTemplate().bind("java:jboss/datasources/euantigenspec", new DriverManagerDataSource("jdbc:h2:mem:testdb"));
    }

    @Test
    void does_not_use_messageId_as_jobId() {
        String messageId = randomData.uuidString();
        final String cpr = randomData.cpr();
        String jobId = coronaPassService.createDataFetcherJob(cpr, messageId, ImmutableMap.of(cpr, KafkaTestUtils.getRandomFetchType()));
        assertNotEquals(messageId, jobId);
    }

    private PersonRecord toPersonRecord(MockDataService.Person person, String cpr) {
        return new PersonRecordImpl()
            .personIdentifier(cpr)
            .givenName(person.first)
            .familyName(person.last);
    }

    @NonNull
    private String cprWithCacheHitOn(@NonNull final String cpr) {
        final Vaccination vaccination = aVaccination(cpr);
        final LabResults labResults = mock(LabResults.class);
        given(cpr, vaccination, labResults);
        return cpr;
    }


    @Test
    void test_getCoronaPass_cache_hit() {
        final String cprPrefix = "040587010";
        for (int i = 0; i < 10; i++) {
            final String cpr = cprWithCacheHitOn(cprPrefix + 1);
            verifyPass(cpr);
        }
    }

    @Test
    void test_get_no_passes_except_dkinfo_case_0405870114() {
        String cpr = "0405870118";

        given(cpr, VaccinationBuilder.vaccination(cpr, false).build(), LabResultsBuilder.cpr(cpr).build());

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertEquals(1, coronaPassResult.getCoronaPasses().size());
        assertEquals(PassType.DKINFO, coronaPassResult.getCoronaPasses().iterator().next().getPassType());
    }

    private EffectuatedVaccination withVaccinationEffectuatedAt(LocalDateTime effectuatedAt) {
        return EffectuatedVaccinationBuilder
            .effectuatedVaccination(32139944147L,
                effectuatedAt).build();
    }

    @Test
    void test_verify_DK_and_verify_DKPlus() {
        String cpr = "0405870113";
        MockDataService.Person person = MockDataService.person(cpr);
        LocalDateTime effectuatedAt = LocalDateTime.now();

        final Vaccination vaccination = vaccinationWithEffectuations(cpr,
            withVaccinationEffectuatedAt(effectuatedAt.minus(2, ChronoUnit.WEEKS)),
            withVaccinationEffectuatedAt(effectuatedAt.minus(4, ChronoUnit.WEEKS)));

        given(cpr, vaccination, noLabResults(cpr));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());

        Cwt dkContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DK);
        Cwt dkPlusContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DKPLUS);

        assertTrue(timestampIsApproxNow(dkContent.getIssuedAt()));
        assertThat(dkContent.getExpiration()).isLessThan(Instant.now().plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(timestampIsApproxNow(dkPlusContent.getIssuedAt()));
        assertThat(dkPlusContent.getExpiration()).isLessThan(Instant.now().plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertEquals("1987-05-04", dkPlusContent.getClaim("dateOfBirth").AsString());
        assertEquals(person.first + " " + person.last, dkPlusContent.getClaim("legalName").AsString());
    }

    private Vaccination aVaccination(String cpr) {
        return VaccinationBuilder.vaccination(cpr, true)
            .add(
                effectuatedVaccination(DRUG_ID,
                    LocalDateTime.now().minus(336, ChronoUnit.HOURS))
                    .build()).build();
    }

    private Vaccination vaccinationWithEffectuations(String cpr, EffectuatedVaccination... effectuatedVaccinations) {
        VaccinationBuilder builder = VaccinationBuilder.vaccination(cpr, true);
        Arrays.stream(effectuatedVaccinations).forEach(builder::add);
        return builder.build();
    }

    private Vaccination noVaccination(String cpr) {
        return VaccinationBuilder.vaccination(cpr, false).build();
    }

    private LabResults noLabResults(String cpr) {
        return LabResultsBuilder.cpr(cpr).build();
    }

    private void given(String cpr, Vaccination vaccination, LabResults labResults) {
        when(dataReader.getData(cpr)).thenReturn(Collections.singleton(new Data(vaccination, labResults, toPersonRecord(MockDataService.person(cpr), cpr))));
    }

    @Test
    void test_positive_test_which_ends_up_as_eu_recovery_pass() {
        String cpr = "0405870115";
        LabResults labresults = Results.positivePcr(cpr, LocalDateTime.now().minusDays(11));
        given(cpr, noVaccination(cpr), labresults);

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        DigitalGreenCertificate dgc = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());
        verifyEuImmunity(labresults, dgc);
    }

    private static void verifyEuImmunity(LabResults labresults, DigitalGreenCertificate dgc) {
        assertNull(dgc.getV());
        assertNull(dgc.getT());
        assertFalse(dgc.getR().isEmpty());
        RecoveryEntry recovery = Iterables.getOnlyElement(dgc.getR());
        LabResult.PositivePcrResult pcrResult = labresults.mostRecentPositivePcrResult();
        assertNotNull(pcrResult);
        assertEquals(pcrResult.samplingDateTime().toLocalDate(), recovery.getFr());
        assertEquals(pcrResult.samplingDateTime().plusDays(11).toLocalDate(), recovery.getDf());
        assertEquals(pcrResult.samplingDateTime().plusDays(180).toLocalDate(), recovery.getDu());
        assertEquals("DK", recovery.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, recovery.getTg());
        assertEquals(VaccineMapper.ISSUER, recovery.getIs());
        assertNotNull(recovery.getCi());
    }

    @Test
    void test_positive_tests_and_only_immunization_eu_passes() {
        String cpr = "0405870115";

        given(cpr, noVaccination(cpr), Results.positivePcr(cpr, LocalDateTime.now().minusHours(14 * 24).minusMinutes(15)));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        DigitalGreenCertificate dgc = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());
        assertFalse(dgc.getR().isEmpty());
        assertNull(dgc.getV());
        assertNull(dgc.getT());
    }


    @Test
    void test_getCoronaPassOnlyEU() {
        String cpr = "0405870115";

        given(cpr, aVaccination(cpr), noLabResults(cpr));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        assertEquals(4, coronaPassResult.getCoronaPasses().size());

        verifyDKContent(coronaPassResult);
        DigitalGreenCertificate vaccineDGC = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());

        String expectedGnt = MrzParser.toMrz(vaccineDGC.getNam().getGn(), -1);
        String expectedFnt = MrzParser.toMrz(vaccineDGC.getNam().getFn(), -1);
        assertEquals(expectedGnt, vaccineDGC.getNam().getGnt());
        assertEquals(expectedFnt, vaccineDGC.getNam().getFnt());

    }

    @Test
    void test_getCoronaPassEUMRZnames() {
        String cpr = cprWithCacheHitOn("2011451636");
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        assertEquals(4, coronaPassResult.getCoronaPasses().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPassResult.getCoronaPasses());
        assertEquals(1, listDGC.size());
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getNam().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getNam().getFn(), -1);
            assertEquals(expectedGnt, certificate.getNam().getGnt());
            assertEquals(expectedFnt, certificate.getNam().getFnt());
        });
    }

    @Test
    void test_getCoronaPass_15_minutes_to_expiry_based_on_PCR_LabResults() {
        String cpr = "0405870117";
        LocalDateTime samplingDateTime = LocalDateTime.now().minus(euPcrMaxAge).plusMinutes(15);
        given(cpr, noVaccination(cpr), Results.negativePcr(cpr, samplingDateTime));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());

        List<Cwt> rawCwts = payloadToRawEUCWT(coronaPassResult.getCoronaPasses());
        assertEquals(1, rawCwts.size());

        rawCwts.stream().peek(c -> System.out.println(c.getExpiration())).forEach(c -> assertTrue(timestampIsApproxNow(c.getExpiration().minus(15, ChronoUnit.MINUTES))));
    }

    @Test
    void test_getCoronaPass_expiry_based_on_vaccination() {
        String cpr = "0405870107";

        LocalDateTime effectuatedAt = LocalDateTime.ofInstant(Instant.now().minus(177 * 24, ChronoUnit.HOURS), ZoneOffset.UTC);
        final Vaccination vaccination = VaccinationBuilder.vaccination(cpr, true).add(effectuatedVaccination(DRUG_ID, effectuatedAt).build()).build();

        given(cpr, vaccination, noLabResults(cpr));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());

        verifyDKContent(coronaPassResult);

        List<Cwt> euCwts = payloadToRawEUCWT(coronaPassResult.getCoronaPasses());
        assertEquals(1, euCwts.size());

        euCwts.forEach(c -> assertThat(c.getExpiration()).isLessThan(Instant.now().plus(euMaximumValidity).plus(1, ChronoUnit.MINUTES)));
    }


    private void verifyPass(String cpr) {
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        final String cpr = "0405870101";

        given(cpr, null, null);
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertFalse(coronaPassResult.isCacheHit());
        assertNull(coronaPassResult.getCoronaPasses());
        assertNotNull(coronaPassResult.getJobId());
    }


    private void verifyDKContent(CoronaPassResult coronaPassResult) {
        Cwt dkContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DK);
        Cwt dkPlusContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DKPLUS);

        assertTrue(dkContent.getExpiration().isAfter(Instant.now()));
        assertThat(dkContent.getExpiration()).isLessThan(Instant.now().plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertTrue(dkContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(dkPlusContent.getExpiration().isAfter(Instant.now()));
        assertThat(dkPlusContent.getExpiration()).isLessThan(Instant.now().plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertTrue(dkPlusContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertNotNull(dkPlusContent.getClaim("dateOfBirth"));
        assertNotNull(dkPlusContent.getClaim("legalName"));
    }

    private static List<DigitalGreenCertificate> payloadToEUpasses(Collection<CoronaPass> passes) {
        return passes.stream().filter(p -> p.getPassType() == PassType.EU).map(CoronaPass::getToken).map(s -> s.substring(4)).map(CwtTestUtils::getDigitalGreenCertificate).collect(Collectors.toList());
    }

    private static List<Cwt> payloadToRawEUCWT(Collection<CoronaPass> passes) {
        return passes.stream().filter(p -> p.getPassType() == PassType.EU).map(CoronaPass::getToken).map(CwtTestUtils::getCwt).collect(Collectors.toList());
    }

    private boolean timestampInRange(Instant start, Instant end, Instant subject) {
        return !subject.isBefore(start) && !subject.isAfter(end);
    }

    private boolean timestampIsApproxNow(Instant subject) {
        return timestampInRange(Instant.now().minus(2, ChronoUnit.SECONDS), Instant.now().plus(2, ChronoUnit.SECONDS), subject);
    }
}
