package dk.sundhedsdatastyrelsen.cbs.service.handler;

import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.service.config.SignatureConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CertificateSigningException;
import dk.sundhedsdatastyrelsen.cbs.service.model.SignedCertificateCWT;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.CountryOfTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.DiseaseAgent;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestSample;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.service.DGCEncoder;

import java.text.MessageFormat;
import java.time.Duration;
import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.when;

@SuppressWarnings("unused") // @ParameterizedTest uses methods and parameters that are not picked up by inspections
@SpringJUnitConfig({SignatureConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@CbsTest
class EUCovidTestCertificateSigningHandlerTest {

    public static final String APPROVED_MANUFACTURER = "1833";
    @Autowired
    private DGCEncoder encoder;

    @Mock
    private EUAntigenFiltration euAntigenFiltration;

    private static final Duration maxValidity = Duration.ofHours(72L);
    private EUCovidTestCertificateSigningHandler signingHandler;
    private static final ZonedDateTime now = ZonedDateTime.now(ZoneId.of("Europe/Copenhagen"));

    @BeforeEach
    void setUp() {
        signingHandler = new EUCovidTestCertificateSigningHandler(encoder, euAntigenFiltration, maxValidity);
        when(euAntigenFiltration.isIdDeviceLegalAtTime(eq(APPROVED_MANUFACTURER), any())).thenReturn(true);
    }

    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("validRequests")
    void should_issue_certificate_for_valid_request(GetSignedEUCovidTestCertificateRequest request, String testName) {
        final SignedCertificateCWT signedCertificateCWT = signingHandler.getSignedCertificateCWT(request, new AuditData());

        assertNotNull(signedCertificateCWT.getCwt());
        assertNotNull(signedCertificateCWT.getCi());
    }

    static Stream<Arguments> validRequests() {
        return Stream.of(
            createRequest(TestType.PCR, TestResult.NEGATIVE, now, null),
            createRequest(TestType.PCR, TestResult.POSITIVE, now, null),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now, APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.POSITIVE, now, APPROVED_MANUFACTURER)
        ).map(request -> Arguments.arguments(request, name(request)));
    }

    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("invalidRequests")
    void should_not_issue_certificate_for_invalid_request(GetSignedEUCovidTestCertificateRequest request, String testName) {
        final AuditData auditData = new AuditData();
        assertThrows(CertificateSigningException.class, () -> signingHandler.getSignedCertificateCWT(request, auditData));
    }

    static Stream<Arguments> invalidRequests() {
        return Stream.of(
            createRequest(TestType.PCR, TestResult.NEGATIVE, now.minus(maxValidity.plusMinutes(5L)), null),
            createRequest(TestType.PCR, TestResult.POSITIVE, now.minus(maxValidity.plusMinutes(5L)), null),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now.minus(maxValidity.plusMinutes(5L)), APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.POSITIVE, now.minus(maxValidity.plusMinutes(5L)), APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.POSITIVE, now, "1234"),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now, "1234")
        ).map(request -> Arguments.arguments(request, name(request)));
    }

    private static String name(GetSignedEUCovidTestCertificateRequest request) {
        return MessageFormat.format("Type: {0} Result: {1} Manufacturer: {2} SampleCollected: {3}",
            request.getPayload().getTypeOfTest(),
            request.getPayload().getTestResult(),
            request.getPayload().getManufacturer(),
            request.getPayload().getSampleCollected());
    }


    private static GetSignedEUCovidTestCertificateRequest createRequest(TestType testType, TestResult testResult, ZonedDateTime sampleCollected, String manufacturer) {
        final GetSignedEUCovidTestCertificateRequest request = new GetSignedEUCovidTestCertificateRequest();
        final TestSample payload = new TestSample();
        payload.setGivenName("Brian");
        payload.setFamilyName("London");
        payload.setDateOfBirth(LocalDate.of(1992, Month.DECEMBER, 17));
        payload.setCountryOfTest(CountryOfTest.DK);
        payload.setDiseaseAgentTargeted(DiseaseAgent.COVID_19);
        payload.setTestingCenter("Billund Lufthavn - Falck A/S");
        payload.setManufacturer(manufacturer);
        payload.setSampleCollected(sampleCollected);
        payload.setTypeOfTest(testType);
        payload.setTestResult(testResult);

        request.setPersonIdentifier("12345678");
        request.setPayload(payload);
        return request;
    }
}
