/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.service.config.DkPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.SignatureConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.VaccineMapperConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.test.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Result;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.signatures.cwt.Cwt;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.List;
import java.util.Set;

import static dk.sundhedsdatastyrelsen.cbs.test.Data.MODERNA_DRUG_IDENTIFIER;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;

@SpringJUnitConfig({DkPassportConfiguration.class, VaccineMapperConfiguration.class, SignatureConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
class DkInfoPassportFactoryTest {
    public static final String CPR = "0506504003";
    public static final String DOB = "1950-06-05";
    public static final PersonRecord PERSON_RECORD = new PersonRecordImpl()
        .personIdentifier(CPR)
        .givenName("John Johnny")
        .familyName("Johnson");

    @Autowired
    private DkInfoPassportFactory dkInfoPassportFactory;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.maximum-period-after-recovery}')}")
    private TemporalAmount maximumPeriodAfterRecovery;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.suspension-period-after-recovery}')}")
    private TemporalAmount suspensionPeriodAfterRecovery;


    @Test
    void audit_logs_DKINFO_token() {
        final Vaccination vaccination = VaccinationBuilder.vaccination(CPR, true)
            .add(EffectuatedVaccinationBuilder.effectuatedVaccination(MODERNA_DRUG_IDENTIFIER, LocalDateTime.now().minusDays(20)).build())
            .build();

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).plusDays(5); //such that positive test within last 5 days doesnt suspend recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final AuditData mockAuditData = mock(AuditData.class);
        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, mockAuditData, null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();
        final String token = coronaPass.getToken();

        verify(mockAuditData).withSensitivePersonalInformation(eq("CWT_DKINFO"), argThat((List<String> arg) -> arg.contains(token)));
    }

    @Test
    void test_full_DKINFO_token() {
        final Vaccination vaccination = VaccinationBuilder.vaccination(CPR, true)
            .add(EffectuatedVaccinationBuilder.effectuatedVaccination(MODERNA_DRUG_IDENTIFIER, LocalDateTime.now().minusDays(20)).build())
            .build();

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).plusDays(5); //such that positive test within last 5 days doesnt suspend recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        System.out.println(token);
        System.out.println();
        System.out.println(cwt);
        System.out.println();

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNotNull(t);
        assertNotNull(v);
        assertNotNull(r);
        assertEquals(1, t.size());
        assertEquals(1, v.size());
        assertEquals(1, r.size());

        final String recoveryValidUntil = r.get(0).get("du").AsString();
        assertEquals(recoveryPcrSamplingDate.plus(maximumPeriodAfterRecovery).toLocalDate().toString(), recoveryValidUntil);

        assertEquals(PERSON_RECORD.familyName(), cwt.getClaim(-260).get(1).get("nam").get("fn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(DOB, cwt.getClaim(-260).get(1).get("dob").AsString());
    }

    @Test
    void test_no_recovery() {
        final Vaccination vaccination = VaccinationBuilder.vaccination(CPR, true)
            .add(EffectuatedVaccinationBuilder.effectuatedVaccination(MODERNA_DRUG_IDENTIFIER, LocalDateTime.now().minusDays(20)).build())
            .build();

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).minusDays(5); //such that positive test within last 5 days suspends recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        System.out.println(token);
        System.out.println();
        System.out.println(cwt);
        System.out.println();

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNotNull(t);
        assertNotNull(v);
        assertNull(r);
        assertEquals(1, t.size());
        assertEquals(1, v.size());
    }


    @Test
    void contains_name_even_when_no_data() {
        final Vaccination noVaccinations = VaccinationBuilder.vaccination(CPR, false).build();
        final LabResults noLabResults = LabResultsBuilder.cpr(CPR).build();
        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(noVaccinations, noLabResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        System.out.println(token);
        System.out.println();
        System.out.println(cwt);
        System.out.println();

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNull(t);
        assertNull(v);
        assertNull(r);

        assertEquals(PERSON_RECORD.familyName(), cwt.getClaim(-260).get(1).get("nam").get("fn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(DOB, cwt.getClaim(-260).get(1).get("dob").AsString());
    }
}
