/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.VaccineMapperException;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import se.digg.dgc.payload.v1.VaccinationEntry;

import java.net.URI;
import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static java.time.LocalDateTime.now;
import static org.junit.jupiter.api.Assertions.*;

class VaccineMapperTest {
    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());
    private VaccineMapper vaccineMapper;
    private VaccineMappingConfiguration vaccineMappingConfiguration;
    private final static long DRUG_ID = 1;
    private final static long DRUG_ID_2 = 2;

    @BeforeEach
    void beforeEach() {
        Map<Long, VaccineMappingConfiguration.VaccineInformation> mappingConfiguration = new HashMap<>();
        mappingConfiguration.put(DRUG_ID, new VaccineMappingConfiguration.VaccineInformation(UUID.randomUUID().toString(), UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        mappingConfiguration.put(DRUG_ID_2, new VaccineMappingConfiguration.VaccineInformation(UUID.randomUUID().toString(), UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        vaccineMappingConfiguration = new VaccineMappingConfiguration(mappingConfiguration);
        vaccineMapper = new VaccineMapper(vaccineMappingConfiguration);
    }

    @Test
    void testIncompleteConfiguration() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(300L, effectuated)
            .build();
        final Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build();

        VaccineMapperException exception = assertThrows(VaccineMapperException.class, () -> vaccineMapper.toVaccinationEntries(vaccination));
        assertTrue(exception.getMessage().contains("Unknown drugIdentifier: '300' - vaccine cannot be mapped"));
    }

    @Test
    void mapCompletedSingleVaccine() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(DRUG_ID, effectuated)
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.toVaccinationEntries(vaccination);

        assertEquals(1, vaccinationEntries.size());

        VaccinationEntry actualVac = vaccinationEntries.get(0);
        assertEquals(vaccination.completed().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(1, actualVac.getDn());
        assertEquals(1, actualVac.getSd());
        assertEquals("Danish Health Data Authority", actualVac.getIs());
        assertDoesNotThrow(() -> new URI(actualVac.getCi()));
    }

    @Test
    void mapCompletedVaccineWithTwoEffectuations() {
        LocalDateTime effectuated = now();
        final LocalDateTime firstEffectuation = effectuated.minusDays(20);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true)
            .add(effectuatedVaccination(DRUG_ID, firstEffectuation)
                .build())
            .add(effectuatedVaccination(DRUG_ID, effectuated)
                .build()).build();

        final VaccineMappingConfiguration.VaccineInformation vaccineIdValues = vaccineMappingConfiguration.getVaccineIdValues(DRUG_ID);

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.toVaccinationEntries(vaccination);

        assertEquals(2, vaccinationEntries.size());

        VaccinationEntry firstVac = vaccinationEntries.get(0);
        assertEquals(firstEffectuation.toLocalDate(), firstVac.getDt());
        assertEquals(vaccineIdValues.ma, firstVac.getMa());
        assertEquals(vaccineIdValues.vp, firstVac.getVp());
        assertEquals(vaccineIdValues.mp, firstVac.getMp());
        assertEquals("DK", firstVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, firstVac.getTg());
        assertEquals(1, firstVac.getDn());
        assertEquals(2, firstVac.getSd());


        VaccinationEntry secondVac = vaccinationEntries.get(1);
        assertEquals(vaccination.completed().toLocalDate(), secondVac.getDt());
        assertEquals(vaccineIdValues.ma, secondVac.getMa());
        assertEquals(vaccineIdValues.vp, secondVac.getVp());
        assertEquals(vaccineIdValues.mp, secondVac.getMp());
        assertEquals("DK", secondVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, secondVac.getTg());
        assertEquals(2, secondVac.getDn());
        assertEquals(2, secondVac.getSd());

    }

    @Test
    void mapIncompleteVaccineWithOneEffectuatedVaccine() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(DRUG_ID, effectuated)
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false).add(effectuatedVaccination).build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.toVaccinationEntries(vaccination);

        assertEquals(1, vaccinationEntries.size());

        VaccinationEntry actualVac = vaccinationEntries.get(0);
        assertEquals(vaccination.mostRecentlyEffectuated().effectuated().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(1, actualVac.getDn());
        assertEquals(2, actualVac.getSd());
    }

    @Test
    @DisplayName("CR-242: Effektueringer fra 'Delvist vaccineret' DDV-svar skal alle have nummerangivelsen '1/2' i EU vac CWTs")
    void mapIncompleteVaccination() {
        final LocalDateTime firstDt = now().minusDays(30);
        final LocalDateTime secondDt = now().minusDays(29);
        final LocalDateTime thirdDt = now().minusDays(10);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false)
            .add(effectuatedVaccination(DRUG_ID, firstDt).build())
            .add(effectuatedVaccination(DRUG_ID, secondDt).build())
            .add(effectuatedVaccination(DRUG_ID_2, thirdDt).build())
            .build();

        final VaccineMappingConfiguration.VaccineInformation vaccine1 = vaccineMappingConfiguration.getVaccineIdValues(DRUG_ID);
        final VaccineMappingConfiguration.VaccineInformation vaccine2 = vaccineMappingConfiguration.getVaccineIdValues(DRUG_ID_2);

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.toVaccinationEntries(vaccination);

        assertEquals(3, vaccinationEntries.size());

        VaccinationEntry firstVac = vaccinationEntries.get(0);
        assertEquals(firstDt.toLocalDate(), firstVac.getDt());
        assertEquals(vaccine1.ma, firstVac.getMa());
        assertEquals(vaccine1.vp, firstVac.getVp());
        assertEquals(vaccine1.mp, firstVac.getMp());
        assertEquals("DK", firstVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, firstVac.getTg());
        assertEquals(1, firstVac.getDn());
        assertEquals(2, firstVac.getSd());

        VaccinationEntry secondVac = vaccinationEntries.get(1);
        assertEquals(secondDt.toLocalDate(), secondVac.getDt());
        assertEquals(vaccine1.ma, secondVac.getMa());
        assertEquals(vaccine1.vp, secondVac.getVp());
        assertEquals(vaccine1.mp, secondVac.getMp());
        assertEquals("DK", secondVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, secondVac.getTg());
        assertEquals(1, secondVac.getDn());
        assertEquals(2, secondVac.getSd());

        VaccinationEntry thirdVac = vaccinationEntries.get(2);
        assertEquals(thirdDt.toLocalDate(), thirdVac.getDt());
        assertEquals(vaccine2.ma, thirdVac.getMa());
        assertEquals(vaccine2.vp, thirdVac.getVp());
        assertEquals(vaccine2.mp, thirdVac.getMp());
        assertEquals("DK", thirdVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, thirdVac.getTg());
        assertEquals(1, thirdVac.getDn());
        assertEquals(2, thirdVac.getSd());
    }

}
