/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.google.common.collect.Range;
import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.annotations.RequiresEmbeddedChildPassesEnabled;
import dk.sundhedsdatastyrelsen.cbs.annotations.RequiresLocalDB;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSDGWSDriver;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSDriver;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSEuSigningDriver;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataCacheDriver;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataCacheInvalidatorDriver;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataFetcherDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver.Citizen;
import dk.sundhedsdatastyrelsen.cbs.kidsdriver.KidsDriver;
import dk.sundhedsdatastyrelsen.cbs.nasdriver.NASDriver;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.util.Wait;
import dk.sundhedsdatastyrelsen.cbs.webservice.ResponseTester;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.CountryOfTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.DiseaseAgent;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestSample;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestType;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import dk.vaccinationsregister.schemas._2013._12._01.DGWSFault;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.TestInstance.Lifecycle;
import org.junit.jupiter.api.extension.ExtendWith;
import org.opentest4j.AssertionFailedError;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.EnabledIf;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.*;
import se.digg.dgc.service.DGCDecoder;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;

import javax.xml.ws.soap.SOAPFaultException;
import java.time.*;
import java.time.temporal.ChronoUnit;
import java.util.Objects;
import java.util.*;
import java.util.stream.Collectors;
import java.util.zip.ZipException;

import static dk.sundhedsdatastyrelsen.cbs.util.CoronaPassAssertions.assertHasName;
import static dk.sundhedsdatastyrelsen.cbs.util.LabResultTester.assertThatTestEntry;
import static dk.sundhedsdatastyrelsen.cbs.util.VaccinationTester.assertThatVaccination;
import static dk.sundhedsdatastyrelsen.cbs.webservice.ResponseTester.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser.*;
import static org.junit.jupiter.api.Assertions.*;

@ExtendWith(SpringExtension.class)
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
@TestInstance(Lifecycle.PER_CLASS)
@DisplayNameGeneration(DisplayNameGenerator.ReplaceUnderscores.class)
class CoronaPassITCase {
    private static final Duration TOKEN_ISSUED_AT_SLACK = Duration.ofSeconds(10L);  // Clocks might be out of sync

    private static final String MANUFACTURER_HSC = "1833";
    private static final String MANUFACTURER_NOT_HSC = "1608";
    private static final String MANUFACTURER_HSC_AND_CE = "1833";
    private static final String MANUFACTURER_CE = "1608";
    private static final String MANUFACTURER_UNKNOWN = "";

    @Autowired
    private DDVDriver ddvDriver;
    @Autowired
    private NASDriver nasDriver;
    @Autowired
    private CBSDriver cbsDriver;
    @Autowired
    private DGCDecoder decoder;
    @Autowired
    private CBSDGWSDriver cbsdgwsDriver;
    @Autowired
    private CBSEuSigningDriver cbsEuSigningDriver;
    @Autowired
    private DataFetcherDriver dataFetcherDriver;
    @Autowired
    private DataCacheInvalidatorDriver dataCacheInvalidatorDriver;
    private Instant now;
    @Autowired
    private KidsDriver kidsDriver;
    @Autowired
    private DataCacheDriver dataCacheDriver;

    private final Queue<IdwsUser> userQueue = new LinkedList<>();

    @BeforeAll
    public void setupAll() throws Exception {
        DigitalGreenCertificate.getCBORMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        nasDriver.setUp();
        userQueue.addAll(Arrays.asList(
            CPR_2401010011,
//            CPR_2401010012,
            CPR_2401010013,
            CPR_2401010014,
            CPR_2401010015,
            CPR_2401010018
        ));
    }

    private IdwsUser nextUser() {
        final IdwsUser user = userQueue.poll();
        userQueue.offer(user);
        System.out.println("CURRENT USER: " + user.cpr());
        return user;
    }

    @AfterAll
    public void tearDownAll() throws Exception {
        nasDriver.tearDown();
    }

    @BeforeEach
    void setUp() {
        now = Instant.now();
    }

    @Nested
    @DisplayName("CBS Acceptance Tests")
    class AcceptanceTests {
        @Test
        void test_A1_when_cache_is_empty_future_is_returned() throws Exception {
            final IdwsUser user = GYLDIGT_TEST_PAS;
            clearDdvAndWaitForInvalidationAndDataFetch(user);
            final GetCoronaPassResponse response = cbsDriver.getCoronaPass(user);
            assertNotNull(response.getCoronaPasses());
            assertNull(response.getFutureResult());

            assertThat(response)
                .hasDkPasses(2);
        }

        @Test
        void test_A2_repeated_calls_returns_freshly_issued_tokens() throws InterruptedException {
            final IdwsUser user = GYLDIGT_TEST_PAS;
            GetCoronaPassResponse response = getCoronaPasses(user);
            final List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            Awaitility.waitAtMost(TOKEN_ISSUED_AT_SLACK.plusSeconds(10L))
                .untilAsserted(() -> assertThrows(AssertionFailedError.class, () -> validateTokensAreValidAndFresh(passList)));

            response = cbsDriver.getCoronaPass(user);
            validateTokensAreValidAndFresh(response.getCoronaPasses().getPass());
        }

        // TODO: 28-06-2021 : Find new way to reliably trigger error in datafetcher
        @Test
        @Disabled("No way to trigger this error anymore.")
        void test_A3_failed_jobs_due_to_integration_error_are_sent_to_NAS() throws InterruptedException {
            final GetCoronaPassResponse futureResult = cbsDriver.getCoronaPass(IdwsUser.NOT_IN_NSP_STAMDATA);
            assertNull(futureResult.getCoronaPasses());
            assertNotNull(futureResult.getFutureResult());
            String jobId = futureResult.getFutureResult().getJobId();
            final JobStatusNotification jobStatusNotification = fetchDataAndWaitForJobStatusNotification(jobId);
            assertEquals(JobStatus.FAILED, jobStatusNotification.getJobStatus());
        }

        @Test
        @RequiresLocalDB
        void test_A5_user_with_only_test_result() throws Exception {
            clearDdvAndWaitForInvalidationAndDataFetch(GYLDIGT_TEST_PAS);
            GetCoronaPassResponse response = getCoronaPasses(GYLDIGT_TEST_PAS);

            List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(0)
                .hasEuLabPasses(1)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A6_user_with_only_vaccination() throws Exception {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;
            Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS)));
            citizen.addPfizerVaccination(Date.from(now.minus(14, ChronoUnit.DAYS)));

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);

            validateTokensAreValidAndFresh(response.getCoronaPasses().getPass());
            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(2)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }


        @Test
        void test_A7_user_with_only_immunity() throws Exception {
            clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.GYLDIGT_IMMUNITET_PAS);
            GetCoronaPassResponse response = getCoronaPasses(IdwsUser.GYLDIGT_IMMUNITET_PAS);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(1);
        }

        @Test
        void test_A8_user_with_too_early_vac_gets_no_DK_passes() throws Exception {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;

            ddvDriver
                .withCitizen(user).clear()
                .addPfizerVaccination(Date.from(now.minus(1, ChronoUnit.DAYS)));

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);

            assertThat(response)
                .hasDkPasses(0)
                .hasEuVacPasses(1)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }


        @Test
        @RequiresLocalDB
        void test_A8_user_with_positive_test_gets_no_passes() throws Exception {
            final IdwsUser user = IdwsUser.CPR_2401010018;
            clearDdvAndWaitForInvalidationAndDataFetch(user);

            kidsDriver
                .withUser(user).clearKids()
                .addPositivePcr(LocalDateTime.now().minusHours(2));
            dataCacheDriver.withUser(user).invalidateLabCache();

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);

            assertThat(response)
                .hasDkPasses(0)
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        // TODO: 18-06-2021 : Depends on modified data in insert-kids-data.sql: Find a cpr from test1 with expired immunity instead
        @Test
        @RequiresLocalDB
        void test_A8_user_with_expired_immunity_gets_no_passes() throws Exception {
            clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.EXPIRED_IMMUNITY);
            GetCoronaPassResponse response = cbsDriver.getCoronaPass(IdwsUser.EXPIRED_IMMUNITY);

            assertThat(response)
                .hasDkPasses(0)
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A9_user_with_no_tests_or_vaccinations_gets_no_passes_returned() throws Exception {
            clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.NO_LAB_RESULTS);
            GetCoronaPassResponse response = getCoronaPasses(IdwsUser.NO_LAB_RESULTS);

            assertThat(response)
                .hasDkPasses(0)
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A10_CPRMismatch() {
            // Create cpr mismatch:
            assertThrows(SOAPFaultException.class, () -> cbsDriver.getCoronaPass(IdwsUser.CPR_3108732076, "1111111118"));
        }

        @Test
        @RequiresLocalDB
            //Doesnt work on test1 due automatic updates of stamdata_person.
        void test_A11_user_not_in_CPR_db() {
            final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> getCoronaPasses(IN_NSP_STAMDATA_BUT_NOT_IN_CPR));
            Truth.assertThat(exception).hasMessageThat().contains("Person data is missing");
        }

        @Test
        @RequiresLocalDB
        @Disabled("Not reliably triggering error")
        void test_A12_failed_jobs_due_to_internal_error_are_sent_to_NAS() throws InterruptedException, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault {
            clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.NO_LAB_RESULTS);
            //trigger CR-198 ("LabResult ETL FAILED" due to primary key violation)
            for (int i = 0; i < 10; i++) {
                final GetCoronaPassResponse futureResult = cbsDriver.getCoronaPass(IdwsUser.NO_LAB_RESULTS);
                assertNotNull(futureResult.getFutureResult());
            }
            final GetCoronaPassResponse futureResult = cbsDriver.getCoronaPass(IdwsUser.NO_LAB_RESULTS);
            assertNotNull(futureResult.getFutureResult());
            final JobStatusNotification jobStatusNotification = fetchDataAndWaitForJobStatusNotification(futureResult.getFutureResult().getJobId());
            assertEquals(JobStatus.FAILED, jobStatusNotification.getJobStatus());
        }

        @Test
        @RequiresLocalDB
        void test_passport_with_all_tokens() throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
            final IdwsUser user = CPR_2401010011;
            Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addModernaVaccination(Date.from(now.minus(60, ChronoUnit.DAYS)));
            citizen.addModernaVaccination(Date.from(now.minus(11, ChronoUnit.DAYS)));

            kidsDriver
                .withUser(user)
                .clearKids()
                .addPositivePcr(LocalDateTime.now().minusDays(150))
                .addNegativePcr(LocalDateTime.now().minusDays(2).minusHours(12))
                .addNegativeAg(LocalDateTime.now().minusDays(2), MANUFACTURER_HSC)
                .addNegativeAg(LocalDateTime.now().minusDays(1), MANUFACTURER_HSC);

            dataCacheDriver
                .withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(2)
                .hasEuLabPasses(3)
                .hasEuRecPasses(1);
        }

        @Test
        @RequiresLocalDB
        void test_DK_passes_gets_removed_after_positive_test() throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
            final IdwsUser user = IdwsUser.CPR_2401010018;
            clearDdvAndWaitForInvalidationAndDataFetch(user);
            kidsDriver
                .withUser(user)
                .clearKids()
                .addNegativeAg(LocalDateTime.now().minusDays(2), MANUFACTURER_HSC);
            dataCacheDriver
                .withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(0)
                .hasEuLabPasses(1)
                .hasEuRecPasses(0);

            kidsDriver
                .withUser(user)
                .addPositiveAg(LocalDateTime.now().minusHours(2), MANUFACTURER_HSC);

            dataCacheDriver
                .withUser(user)
                .invalidateLabCache();

            response = getCoronaPassesAfterInvalidation(user);
            passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(0)
                .hasEuVacPasses(0)
                .hasEuLabPasses(1)
                .hasEuRecPasses(0);
        }

        @Test
        void test_parent_requests_child_passport() throws Exception {
            final IdwsUser parent = IdwsUser.PARENT_A;
            final IdwsUser child = IdwsUser.CHILD_OF_A;

            Citizen citizen = ddvDriver.withCitizen(child.cpr()).clear();
            citizen.addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS)));
            citizen.addPfizerVaccination(Date.from(now.minus(14, ChronoUnit.DAYS)));

            GetCoronaPassResponse response = waitForInvalidationAndGetFutureResult(parent, child);
            assertNull(response.getCoronaPasses());
            assertNotNull(response.getFutureResult());

            String jobId = response.getFutureResult().getJobId();
            fetchDataAndWaitForJobStatusNotification(jobId);
            GetCoronaPassResponse passport = cbsDriver.getCoronaPass(parent, child.cpr());

            validateTokensAreValidAndFresh(passport.getCoronaPasses().getPass());
            ResponseTester.assertThat(passport)
                .hasDkPasses(2)
                .hasEuVacPasses(2);
        }

        @Test
        @DisplayName("Get passport for user with EU-valid test data in little KIDS")
        @RequiresLocalDB
        void checkEUPassForEUValidTest() throws InterruptedException {
            final GetCoronaPassResponse coronaPasses = getCoronaPasses(GYLDIGT_TEST_PAS); //TODO should use a user that has a antigen result with invalid manufacturer, instead of a PCR result
            assertThat(coronaPasses)
                .hasEuLabPasses(1);
        }

        @Test
        @DisplayName("Get passport for user with EU-invalid test data in little KIDS cases no EU test passport")
        void checkNoEUPassForEUInvalidTest() throws InterruptedException {
            final GetCoronaPassResponse coronaPasses = getCoronaPasses(IdwsUser.UNAPPROVED_LAB_RESULT); // has test with idDevice code system prefix "118.", which is not EU-valid
            assertThat(coronaPasses)
                .hasNoEuLabPasses();
        }

        @Test
        @DisplayName("CR-182 - All Certificates (except DK) contains legal name of person")
        @RequiresLocalDB
        void checkCorrectNameInPasses() throws Exception {
            final IdwsUser user = IdwsUser.CPR_2401010012;
            Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addModernaVaccination(Date.from(now.minus(60, ChronoUnit.DAYS)));

            final String givenName = "given name";
            final String family = "family";

            kidsDriver
                .withUser(user)
                .clearKids()
                .addPositivePcr(LocalDateTime.now().minusDays(150))
                .addNegativePcr(LocalDateTime.now().minusDays(2).minusHours(12));

            dataCacheDriver
                .withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            GetCoronaPassResponse response = getCoronaPassesAfterInvalidation(user);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses(2)
                .hasEuVacPasses(1)
                .hasEuLabPasses(1)
                .hasEuRecPasses(1);

            CwtTestUtils.getEuLabPasses(response)
                .stream().map(PassJWT::getJWT)
                .map(CoronaPassITCase.this::getEUPayload)
                .forEach(assertHasName(givenName, family));

            CwtTestUtils.getEuVacPasses(response)
                .stream().map(PassJWT::getJWT)
                .map(CoronaPassITCase.this::getEUPayload)
                .forEach(assertHasName(givenName, family));

            CwtTestUtils.getEuRecPasses(response)
                .stream().map(PassJWT::getJWT)
                .map(CoronaPassITCase.this::getEUPayload)
                .forEach(assertHasName(givenName, family));

            final DigitalGreenCertificate dkInfo = getDKINFOPayload(response);
            assertHasName(givenName, family).accept(dkInfo);

            final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(response);

            final Cwt cwt = getCwtFromToken(dkPlusPass.getJWT().substring(4));
            final String legalName = givenName + " " + family;
            assertEquals(legalName, cwt.getClaim("legalName").AsString());
        }

        @Test
        @DisplayName("CR-182 - DKINFO Pass contains legal name of person")
        void checkCorrectNameInDKINFO() throws Exception {
            final String givenName = "Jakob Jønsson";
            final String family = "Møller";

            GetCoronaPassResponse response = getCoronaPasses(CPR_2401010011);

            final DigitalGreenCertificate dkInfo = getDKINFOPayload(response);
            assertHasName(givenName, family).accept(dkInfo);
        }

        @Test
        @DisplayName("Special Case - All tests are returned in DKINFO, but only approved are returned in EU")
        @RequiresLocalDB
        void checkDKINFOContainsAllTests() throws InterruptedException {
            final IdwsUser user = nextUser();
            dataCacheDriver.withUser(user)
                .invalidateLabCache();

            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(Duration.ofHours(70L));
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(Duration.ofHours(60L));
            final LocalDateTime test3SampleTime = LocalDateTime.now().minus(Duration.ofHours(50L));

            kidsDriver.withUser(user).clearKids()
                .addNegativeAg(test1SampleTime, MANUFACTURER_CE)
                .addNegativeAg(test2SampleTime, MANUFACTURER_HSC)
                .addNegativePcr(test3SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasDkPasses(2)
                .hasEuLabPasses(2);

            final List<TestEntry> testEntriesFromDKINFO = getTestEntriesOrderedBySampleTime(getDKINFOPayload(coronaPasses));

            assertEquals(3, testEntriesFromDKINFO.size());

            assertThatTestEntry(testEntriesFromDKINFO.get(0))
                .isNegative()
                .isAG()
                .hasSampleTime(test1SampleTime)
                .hasManufacturer(MANUFACTURER_CE);

            assertThatTestEntry(testEntriesFromDKINFO.get(1))
                .isNegative()
                .isAG()
                .hasSampleTime(test2SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            assertThatTestEntry(testEntriesFromDKINFO.get(2))
                .isNegative()
                .isPcr()
                .hasSampleTime(test3SampleTime);

            final List<TestEntry> testEntriesFromEU = getTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(coronaPasses));

            assertEquals(2, testEntriesFromEU.size());

            assertThatTestEntry(testEntriesFromEU.get(0))
                .isNegative()
                .isAG()
                .hasSampleTime(test2SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            assertThatTestEntry(testEntriesFromEU.get(1))
                .isNegative()
                .isPcr()
                .hasSampleTime(test3SampleTime);
        }

        @Test
        @DisplayName("CR-194.1 - PCR with longer validity")
        @RequiresLocalDB
        void longestValidTestDeterminesCertificateValidity() throws Exception {
            final IdwsUser user = nextUser();

            ddvDriver.withCitizen(user).clear();

            dataCacheDriver.withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            final LocalDateTime time = LocalDateTime.now();
            final LocalDateTime test1SampleTime = time.minus(Duration.ofHours(95L));
            final LocalDateTime test2SampleTime = time.minus(Duration.ofHours(73L));

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasDkPasses(2)
                .hasEuLabPasses(2);

            final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

            final Cwt cwt = getCwtFromToken(dkPlusPass.getJWT().substring(4));

            final Duration expiresIn = Duration.between(time.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), cwt.getExpiration());

            Truth.assertThat(expiresIn)
                .isIn(Range.closed(Duration.ofHours(1).minusSeconds(60L), Duration.ofHours(1)));
        }

        @Test
        @DisplayName("CR-194.2 - No DK pass if test certificates are too old")
        @RequiresLocalDB
        void expiredTests() throws Exception {
            final IdwsUser user = nextUser();

            ddvDriver.withCitizen(user).clear();

            dataCacheDriver.withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            final LocalDateTime time = LocalDateTime.now();
            final LocalDateTime test1SampleTime = time.minus(Duration.ofHours(97L));
            final LocalDateTime test2SampleTime = time.minus(Duration.ofHours(73L));

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasNoDkPasses()
                .hasEuLabPasses(2);
        }

        @Test
        @DisplayName("CR-194.3 - PCR with longer validity determines token expiry")
        @RequiresLocalDB
        void longestValidTestDeterminesCertificateValidityPCRLongest() throws Exception {
            final IdwsUser user = nextUser();

            ddvDriver.withCitizen(user).clear();

            dataCacheDriver.withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            final LocalDateTime time = LocalDateTime.now();
            final LocalDateTime test1SampleTime = time.minus(Duration.ofHours(95L).plusMinutes(15L)); // valid for 45 minutes
            final LocalDateTime test2SampleTime = time.minus(Duration.ofHours(71L).plusMinutes(30L)); // valid for 30 minutes

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasDkPasses(2)
                .hasEuLabPasses(2);

            final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

            final Cwt cwt = getCwtFromToken(dkPlusPass.getJWT().substring(4));

            final Duration expiresIn = Duration.between(time.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), cwt.getExpiration());

            Truth.assertThat(expiresIn)
                .isIn(Range.closed(Duration.ofMinutes(44), Duration.ofMinutes(45)));
        }

        @Test
        @DisplayName("CR-194.4 - AG test with longer validity determines token expiry")
        @RequiresLocalDB
        void longestValidTestDeterminesCertificateValidityAgLongest() throws Exception {
            final IdwsUser user = nextUser();

            ddvDriver.withCitizen(user).clear();

            dataCacheDriver.withUser(user)
                .invalidateLabCache()
                .invalidateVacCache();

            final LocalDateTime time = LocalDateTime.now();
            final LocalDateTime test1SampleTime = time.minus(Duration.ofHours(95L).plusMinutes(55L)); // valid for 5 minutes
            final LocalDateTime test2SampleTime = time.minus(Duration.ofHours(71L).plusMinutes(20L)); // valid for 40 minutes

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasDkPasses(2)
                .hasEuLabPasses(2);

            getTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(coronaPasses));

            final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

            final Cwt cwt = getCwtFromToken(dkPlusPass.getJWT().substring(4));

            final Duration expiresIn = Duration.between(time.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), cwt.getExpiration());

            Truth.assertThat(expiresIn)
                .isIn(Range.closed(Duration.ofMinutes(39), Duration.ofMinutes(40)));
        }


    }


    @Nested
    @DisplayName("EU Vaccination Corona Pass Rules")
    class VaccinationTests {
        final Duration VAC_OFFSET = Duration.ofMinutes(1);
        final Duration VAC_VALID_START = Duration.ofDays(14);
        final Period VAC_VALID_END_COMPLETED = Period.ofMonths(8);
        final Duration VAC_VALID_END_FIRST_EFFECTUATION = Duration.ofDays(42);

        @Test
        @DisplayName("1.1.1 - First effectuation with AstraZeneca at first valid time")
        void case_1_1_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_OFFSET); // FIXME: 16-06-2021 spec says plus, but that is an error, right?

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.2 - First effectuation with AstraZeneca at latest valid time")
        void case_1_1_2() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION.minus(VAC_OFFSET));

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.3 - First effectuation with AstraZeneca at latest valid time threshold ")
        void case_1_1_3() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.4 - First effectuation with AstraZeneca at latest valid time + 1 minute")
        void case_1_1_4() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION.plus(VAC_OFFSET));

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.2.1 - Completed Vaccination with Johnson & Johnson at first valid time")
        void case_1_2_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_START);

            ddvDriver.withCitizen(user)
                .clear()
                .addJohnsonVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime)
                .isJohnsonVaccination();
        }

        @Test
        @DisplayName("1.2.2 - Completed Vaccination with Johnson & Johnson at expiration time")
        @Disabled("It is not possible to create a covid19 vaccination before end December 2020 in DDV")
        void case_1_2_2() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.atZone(ZoneId.of("Europe/Copenhagen")).minus(VAC_VALID_END_COMPLETED).toInstant(); // FIXME: 16-06-2021 spec says 2xVAC_VALID_END_COMPLETED, must be an error?

            ddvDriver.withCitizen(user)
                .clear()
                .addJohnsonVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime)
                .isJohnsonVaccination();
        }

        @Test
        @DisplayName("1.3.1 - Completed Vaccination with AstraZeneca at first valid time")
        void case_1_3_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addAstraZenecaVaccination(Date.from(firstEffectuation));
            citizen.addAstraZenecaVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isAstraZenicaVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.4.1 - Completed Vaccination with AstraZeneca and Moderna")
        void case_1_4_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addAstraZenecaVaccination(Date.from(firstEffectuation));
            citizen.addModernaVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isAstraZenicaVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isModernaVaccination();
        }

        @Test
        @DisplayName("1.5.1 - 2 Vaccinations too close to each other to be considered completed")
        void case_1_5_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_START.plus(VAC_OFFSET));
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addPfizerVaccination(Date.from(firstEffectuation));
            citizen.addPfizerVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isPfizerVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("1.5.2 - 2 Vaccinations too far from each other")
        void case_1_5_2() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.atZone(ZoneId.of("Europe/Copenhagen")).minus(VAC_VALID_END_COMPLETED).toInstant();
            final Instant secondEffectuation = now.minus(VAC_OFFSET);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addPfizerVaccination(Date.from(firstEffectuation));
            citizen.addPfizerVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isPfizerVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isPfizerVaccination();
        }

    }

    @Nested
    @DisplayName("EU Test Corona Pass Rules")
    @RequiresLocalDB
    class LabResultTests {
        final Duration TEST_OFFSET = Duration.ofMinutes(1L);
        final Period TEST_DISPLAY_MAX = Period.ofWeeks(1);// Duration.ofDays(7L); // FIXME: 17-06-2021 : spec says 7 days, but config is 1 week
        final Duration TEST_VALID_END = Duration.ofHours(72L);

        @Test
        @DisplayName("2.1.1 - Test certificate immediately after test")
        void case_2_1_1() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_OFFSET);

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(1, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertEquals(1, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);

            assertThatTestEntry(firstTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1SampleTime);
        }

        @Test
        @DisplayName("2.1.2 - Test certificate 7 days minus 1 minute after test")
        void case_2_1_2() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET);

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(1, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertEquals(1, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);

            assertThatTestEntry(firstTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1SampleTime);
        }

        @Test
        @DisplayName("2.1.3 - Test certificate 7 days after test")
        void case_2_1_3() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX);

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasNoEuLabPasses();
        }

        @Test
        @DisplayName("2.1.4 - Test certificate 7 days plus 1 minute after test")
        void case_2_1_4() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).minus(TEST_OFFSET);

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses).hasNoEuLabPasses();
        }

        @Test
        @DisplayName("2.2.1 - Test certificate for negative PCR and approved AG tests and filtering of unapproved negative AG test and positive PCR and AG tests ")
        void case_2_2_1() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(1));
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(2));
            final LocalDateTime test3SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(3));
            final LocalDateTime test4SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(4));
            final LocalDateTime test5SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(5));
            final LocalDateTime test6SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).plus(TEST_OFFSET.multipliedBy(6));

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addPositivePcr(test2SampleTime)
                .addNegativePcr(test3SampleTime)
                .addNegativeAg(test4SampleTime, MANUFACTURER_HSC)
                .addPositiveAg(test5SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test6SampleTime, MANUFACTURER_NOT_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(3, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertEquals(3, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);
            final TestEntry secondTest = testEntries.get(1);
            final TestEntry thirdTest = testEntries.get(2);

            assertThatTestEntry(firstTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1SampleTime);

            assertThatTestEntry(secondTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test3SampleTime);

            assertThatTestEntry(thirdTest)
                .isNegative()
                .isAG()
                .hasSampleTime(test4SampleTime);
        }

        @Test
        @DisplayName("2.3.1 - Test certificate approved AG tests manufacturers")
        void case_2_3_1() throws InterruptedException {
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(0));
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(1));
            final LocalDateTime test3SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(2));
            final LocalDateTime test4SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(3));
            final LocalDateTime test5SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(4));
            final LocalDateTime test6SampleTime = LocalDateTime.now().minus(TEST_VALID_END).plus(TEST_OFFSET.multipliedBy(5));

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user).clearKids()
                .addNegativeAg(test1SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test2SampleTime, MANUFACTURER_NOT_HSC)
                .addNegativeAg(test3SampleTime, MANUFACTURER_HSC_AND_CE)
                .addNegativeAg(test4SampleTime, MANUFACTURER_CE)
                .addNegativeAg(test5SampleTime, MANUFACTURER_UNKNOWN)
                .addNegativeAg(test6SampleTime, "");

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(2, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertEquals(2, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);
            final TestEntry secondTest = testEntries.get(1);

            assertThatTestEntry(firstTest)
                .isNegative()
                .isAG()
                .hasSampleTime(test1SampleTime);

            assertThatTestEntry(secondTest)
                .isNegative()
                .isAG()
                .hasSampleTime(test3SampleTime);
        }


    }

    @Nested
    @DisplayName("EU Recovery Corona Pass Rules")
    @RequiresLocalDB
    class RecoveryTests {

        final Duration RECOVERY_VALID_START = Duration.ofDays(11L);
        final Duration RECOVERY_VALID_END = Duration.ofDays(180L);
        final Duration RECOVERY_OFFSET = Duration.ofMinutes(1L);
        final Duration RECOVERY_POSITIVE_OK = Duration.ofDays(84L);

        @Test
        @DisplayName("3.1.1 - Recovery after positive PCR, invalid time")
        void case_3_1_1() throws InterruptedException {

            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START).plus(RECOVERY_OFFSET);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses).hasNoEuRecPass();
        }

        @Test
        @DisplayName("3.1.2 - Recovery after positive PCR, first valid time")
        void case_3_1_2() throws InterruptedException {

            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.1.3 - Recovery after positive PCR, other valid time")
        void case_3_1_3() throws InterruptedException {

            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START).minus(RECOVERY_OFFSET);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.1.4 - Recovery after positive PCR, last valid time")
        void case_3_1_4() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_END).plus(RECOVERY_OFFSET);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.1.5 - Recovery after positive PCR, first invalid time")
        void case_3_1_5() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_END);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses).hasNoEuRecPass();
        }

        @Test
        @DisplayName("3.1.6 - Recovery after positive PCR, other invalid time")
        void case_3_1_6() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_END).minus(RECOVERY_OFFSET);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses).hasNoEuRecPass();
        }

        @Test
        @DisplayName("3.2.1 - Positive PCR now in Recovery period")
        void case_3_2_1() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_POSITIVE_OK);
            final LocalDateTime test2SampleTime = LocalDateTime.now();

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime)
                .addPositivePcr(test2SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.2.2 - Positive PCR 11 days into Recovery period")
        void case_3_2_2() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_POSITIVE_OK).minus(RECOVERY_VALID_START);
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime)
                .addPositivePcr(test2SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test2SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.2.3 - Positive AG 11 days into Recovery period")
        void case_3_2_3() throws InterruptedException {
            final IdwsUser user = nextUser();
            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_POSITIVE_OK).minus(RECOVERY_VALID_START);
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START);

            dataCacheDriver.withUser(user).invalidateLabCache();

            kidsDriver.withUser(user)
                .clearKids()
                .addPositivePcr(test1SampleTime)
                .addPositiveAg(test2SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());

            final RecoveryEntry recoveryEntry = getRecoveryEntry(euRecPasses);

            assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
        }

        private RecoveryEntry getRecoveryEntry(List<PassJWT> euRecPasses) {
            return euRecPasses.stream()
                .map(PassJWT::getJWT)
                .map(CoronaPassITCase.this::getEUPayload)
                .map(Eudgc::getR)
                .flatMap(List::stream)
                .findFirst()
                .orElseThrow(IllegalStateException::new);
        }
    }

    @Nested
    @DisplayName("Child Pass Cases")
    @RequiresEmbeddedChildPassesEnabled
    @RequiresLocalDB
    class ChildPasses {
        @Value("${cwt.children.include-in-response.dk-passes}")
        private boolean makeChildrenDkPasses;

        @Test
        @DisplayName("243.1 - Child and Parent has Test Certificates")
        void childAndParentHasPassport() throws InterruptedException {
            final IdwsUser user = PARENT_A;
            final IdwsUser child = CHILD_OF_A;

            final LocalDateTime parentSampleTime = LocalDateTime.now().minusHours(24L);
            final LocalDateTime childSampleTime = LocalDateTime.now().minusHours(12L);

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(parentSampleTime);

            kidsDriver.withUser(child).clearKids()
                .addNegativeAg(childSampleTime, MANUFACTURER_HSC);

            dataCacheDriver.withUser(user).invalidateLabCache();
            dataCacheDriver.withUser(child).invalidateLabCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass, decoder)
                .forCustody("Parent",
                    parent -> parent
                        .hasDkPasses(2)
                        .hasDkInfoWithName(user.commonName())
                        .hasEuLabPassWithSampleTime(parentSampleTime)
                        .allCwtsWithNameHasName(user.commonName())

                )
                .forCustody("Child1",
                    child1 -> {
                        if (makeChildrenDkPasses)
                            child1.hasDkPasses(2);
                        else
                            child1.hasNoDkPasses();
                        child1.hasDkInfoWithName(child.commonName());
                        child1.hasEuLabPassWithSampleTime(childSampleTime);
                        child1.allCwtsWithNameHasName(child.commonName());
                    }
                );

            final List<PassJWT> parentPasses = CwtTestUtils.getPassesForCustody(coronaPass, "Parent");
            final List<TestEntry> parentTestEntries = getTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(parentPasses));

            assertEquals(1, parentTestEntries.size());

            final List<PassJWT> childPasses = CwtTestUtils.getPassesForCustody(coronaPass, "Child1");
            final List<TestEntry> childTestEntries = getTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(childPasses));

            assertEquals(1, childTestEntries.size());

            assertThatTestEntry(parentTestEntries.get(0))
                .isPcr()
                .isNegative()
                .hasSampleTime(parentSampleTime);

            assertThatTestEntry(childTestEntries.get(0))
                .isAG()
                .isNegative()
                .hasSampleTime(childSampleTime)
                .hasManufacturer(MANUFACTURER_HSC);
        }

        @Test
        @DisplayName("243.2 - Parent has Test Certificate, Child does not")
        void onlyParentHasPassport() throws InterruptedException {
            final IdwsUser user = PARENT_A;
            final IdwsUser child = CHILD_OF_A;

            final LocalDateTime parentSampleTime = LocalDateTime.now().minusHours(24L);

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(parentSampleTime);

            kidsDriver.withUser(child).clearKids();

            dataCacheDriver.withUser(user).invalidateLabCache();
            dataCacheDriver.withUser(child).invalidateLabCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass)
                .forCustody("Parent", parentPart ->
                    parentPart
                        .hasEuLabPasses(1)
                        .hasDkPasses(2)
                )
                .forCustody("Child1",
                    ResponseTester::hasNoEuLabPasses
                );
        }

        @Test
        @DisplayName("243.3 - Child has Test Certificate, Parent does not")
        void onlyChildHasPassport() throws InterruptedException {
            final IdwsUser user = PARENT_A;
            final IdwsUser child = CHILD_OF_A;

            kidsDriver.withUser(user).clearKids();

            kidsDriver.withUser(child).clearKids()
                .addNegativePcr(LocalDateTime.now().minusHours(24L));

            dataCacheDriver.withUser(user).invalidateLabCache();
            dataCacheDriver.withUser(child).invalidateLabCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass)
                .forCustody("Parent",
                    ResponseTester::hasNoEuLabPasses
                )
                .forCustody("Child1",
                    childPart -> childPart.hasEuLabPasses(1)
                );
        }

        @Test
        @DisplayName("243.4 - Child has Test Certificate, Parent has Vaccination Certificate")
        void childHasTestParentHasVac() throws Exception {
            final IdwsUser user = PARENT_A;
            final IdwsUser child = CHILD_OF_A;

            ddvDriver.withCitizen(user).clear()
                .addPfizerVaccination(Date.from(now.minus(Duration.ofDays(20L))));

            ddvDriver.withCitizen(child).clear();

            kidsDriver.withUser(user).clearKids();

            kidsDriver.withUser(child).clearKids()
                .addNegativePcr(LocalDateTime.now().minusHours(24L));

            dataCacheDriver.withUser(user).invalidateLabCache().invalidateVacCache();
            dataCacheDriver.withUser(child).invalidateLabCache().invalidateVacCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass)
                .forCustody("Parent", parentPart -> parentPart
                    .hasNoEuLabPasses()
                    .hasNoEuRecPass()
                    .hasEuVacPasses(1)
                )
                .forCustody("Child1", childPart -> childPart
                    .hasEuLabPasses(1)
                    .hasNoEuRecPass()
                    .hasNoEuVacPasses()
                );
        }

        @Test
        @DisplayName("243.5 - Child has Recovery Certificate, Parent has Vaccination Certificate and Test Certificate")
        void childHasRecoveryParentHasVacAndTest() throws Exception {
            final IdwsUser user = PARENT_A;
            final IdwsUser child = CHILD_OF_A;

            ddvDriver.withCitizen(user).clear()
                .addPfizerVaccination(Date.from(now.minus(Duration.ofDays(20L))));

            ddvDriver.withCitizen(child).clear();

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(LocalDateTime.now().minusHours(8L));

            kidsDriver.withUser(child).clearKids()
                .addPositivePcr(LocalDateTime.now().minusDays(20L));

            dataCacheDriver.withUser(user).invalidateLabCache().invalidateVacCache();
            dataCacheDriver.withUser(child).invalidateLabCache().invalidateVacCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass)
                .forCustody("Parent", parentPart -> parentPart
                    .hasEuLabPasses(1)
                    .hasNoEuRecPass()
                    .hasEuVacPasses(1)
                )
                .forCustody("Child1", childPart -> childPart
                    .hasNoEuLabPasses()
                    .hasEuRecPasses(1)
                    .hasNoEuVacPasses()
                );
        }

        @Test
        @DisplayName("243.6 - Multiple children with different certificates")
        void multipleChildren() throws Exception {
            final IdwsUser user = PARENT_B;
            final IdwsUser child1 = CHILD_1_OF_B;
            final IdwsUser child2 = CHILD_2_OF_B;
            final IdwsUser child3 = CHILD_3_OF_B;

            ddvDriver.withCitizen(user).clear()
                .addPfizerVaccination(Date.from(now.minus(Duration.ofDays(20L))));

            ddvDriver.withCitizen(child1).clear();
            ddvDriver.withCitizen(child2).clear();
            ddvDriver.withCitizen(child3).clear();

            kidsDriver.withUser(user).clearKids();

            kidsDriver.withUser(child1).clearKids()
                .addPositivePcr(LocalDateTime.now().minusDays(20L));

            kidsDriver.withUser(child2).clearKids()
                .addNegativePcr(LocalDateTime.now().minusHours(5L));

            kidsDriver.withUser(child3).clearKids();

            dataCacheDriver.withUser(user).invalidateLabCache().invalidateVacCache();
            dataCacheDriver.withUser(child1).invalidateLabCache().invalidateVacCache();
            dataCacheDriver.withUser(child2).invalidateLabCache().invalidateVacCache();
            dataCacheDriver.withUser(child3).invalidateLabCache().invalidateVacCache();

            final GetCoronaPassResponse coronaPass = getCoronaPassesAfterInvalidation(user);

            final Map<IdwsUser, String> custodyLabels = custodyLabels(coronaPass);

            assertThat(coronaPass, decoder)
                .forCustody("Parent", parentPart -> parentPart
                    .hasDkPasses(2)
                    .hasDkInfoWithName(user.commonName())
                    .hasNoEuLabPasses()
                    .hasNoEuRecPass()
                    .hasEuVacPasses(1)
                    .allCwtsWithNameHasName(user.commonName())
                )
                .forCustody(custodyLabels.get(child1), childPart -> {
                        if (makeChildrenDkPasses)
                            childPart.hasDkPasses(2);
                        else
                            childPart.hasNoDkPasses();
                        childPart
                            .hasDkInfoWithName(child1.commonName())
                            .hasNoEuLabPasses()
                            .hasEuRecPasses(1)
                            .hasNoEuVacPasses()
                            .allCwtsWithNameHasName(child1.commonName());
                    }
                )
                .forCustody(custodyLabels.get(child2), childPart -> {
                        if (makeChildrenDkPasses)
                            childPart.hasDkPasses(2);
                        else
                            childPart.hasNoDkPasses();
                        childPart
                            .hasDkInfoWithName(child2.commonName())
                            .hasEuLabPasses(1)
                            .hasNoEuRecPass()
                            .hasNoEuVacPasses()
                            .allCwtsWithNameHasName(child2.commonName());
                    }
                )
                .forCustody(custodyLabels.get(child3), childPart -> {
                        childPart
                            .hasNoDkPasses()
                            .hasDkInfoWithName(child3.commonName())
                            .hasNoEuLabPasses()
                            .hasNoEuLabPasses()
                            .hasNoEuVacPasses();

                    }
                );
        }

        @Test
        @DisplayName("243.7 - Children older than 15 are not included in Parents response")
        void childOver15NotIncluded() throws InterruptedException {

            final IdwsUser parent = CPR_2401010011; //TODO: find better test person
            final IdwsUser child = CPR_2401010013; //TODO: find better test person

            kidsDriver.withUser(parent)
                .clearKids()
                .addNegativePcr(LocalDateTime.now());

            kidsDriver.withUser(child)
                .clearKids()
                .addNegativePcr(LocalDateTime.now());

            dataCacheDriver.withUser(parent).invalidateLabCache();
            dataCacheDriver.withUser(child).invalidateLabCache();

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(parent);

            assertThat(coronaPasses).hasNoCustodyChildPasses();
        }

        private Map<IdwsUser, String> custodyLabels(GetCoronaPassResponse coronaPass) {
            return CwtTestUtils.getDKInfoPass(coronaPass).stream()
                .collect(Collectors.toMap(passJWT -> IdwsUser.fromName(getName(getDKINFOPayload(passJWT.getJWT()))), CwtTestUtils::getCustody));
        }

        private String getName(DigitalGreenCertificate certificate) {
            return certificate.getNam().getGn() + " " + certificate.getNam().getFn();
        }
    }

    @Nested
    @DisplayName("EU Corona Pass Rules Special Cases")
    @RequiresLocalDB
    class SpecialCases {

        @Test
        @DisplayName("Special Case - Issue Test Certificates for positive test results")
        @EnabledIf(expression = "${cwt.eu.include-positive-results}",
            reason = "Requires that positive test certificates are enabled on CBS",
            loadContext = true)
        void issue_test_certificates_for_positive_results() throws Exception {

            final IdwsUser user = nextUser();

            dataCacheDriver.withUser(user).invalidateLabCache();

            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(Duration.ofHours(96L));
            final LocalDateTime test2SampleTime = LocalDateTime.now().minus(Duration.ofHours(80L));
            final LocalDateTime test3SampleTime = LocalDateTime.now().minus(Duration.ofHours(72L));

            kidsDriver.withUser(user).clearKids()
                .addPositiveAg(test1SampleTime, MANUFACTURER_HSC)
                .addNegativePcr(test2SampleTime)
                .addPositivePcr(test3SampleTime);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(3, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertThatTestEntry(testEntries.get(0))
                .isAG()
                .isPositive()
                .hasSampleTime(test1SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            assertThatTestEntry(testEntries.get(1))
                .isPcr()
                .isNegative()
                .hasSampleTime(test2SampleTime);

            assertThatTestEntry(testEntries.get(2))
                .isPcr()
                .isPositive()
                .hasSampleTime(test3SampleTime);
        }

        @Test
        @DisplayName("Special Case - Issue at most 5 test certificates of each type within the last 7 days")
        void max_5_test_passes_for_each_type() throws InterruptedException {

            final IdwsUser user = nextUser();
            dataCacheDriver.withUser(user).invalidateLabCache();

            final LocalDateTime test1SampleTime = LocalDateTime.now().minus(Duration.ofHours(72L));
            final LocalDateTime test2SampleTime = test1SampleTime.minusMinutes(1);
            final LocalDateTime test3SampleTime = test1SampleTime.minusMinutes(2);
            final LocalDateTime test4SampleTime = test1SampleTime.minusMinutes(3);
            final LocalDateTime test5SampleTime = test1SampleTime.minusMinutes(4);
            final LocalDateTime test6SampleTime = test1SampleTime.minusMinutes(5);
            final LocalDateTime test7SampleTime = test1SampleTime.minusMinutes(6);
            final LocalDateTime test8SampleTime = test1SampleTime.minusMinutes(7);
            final LocalDateTime test9SampleTime = test1SampleTime.minusMinutes(8);
            final LocalDateTime test10SampleTime = test1SampleTime.minusMinutes(9);
            final LocalDateTime test11SampleTime = test1SampleTime.minusMinutes(10);
            final LocalDateTime test12SampleTime = test1SampleTime.minusMinutes(11);
            final LocalDateTime test13SampleTime = test1SampleTime.minusMinutes(12);
            final LocalDateTime test14SampleTime = test1SampleTime.minusMinutes(13);
            final LocalDateTime test15SampleTime = test1SampleTime.minusMinutes(14);
            final LocalDateTime test16SampleTime = test1SampleTime.minusMinutes(15);

            kidsDriver.withUser(user).clearKids()
                .addNegativePcr(test1SampleTime)
                .addNegativePcr(test2SampleTime)
                .addPositivePcr(test3SampleTime)
                .addNegativePcr(test4SampleTime)
                .addNegativePcr(test5SampleTime)
                .addNegativePcr(test6SampleTime)
                .addNegativePcr(test7SampleTime)
                .addNegativeAg(test8SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test9SampleTime, MANUFACTURER_HSC)
                .addPositiveAg(test10SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test11SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test12SampleTime, MANUFACTURER_CE)
                .addNegativeAg(test13SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test14SampleTime, MANUFACTURER_UNKNOWN)
                .addNegativeAg(test15SampleTime, MANUFACTURER_HSC)
                .addNegativeAg(test16SampleTime, MANUFACTURER_HSC);

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);

            assertEquals(10, euLabPasses.size());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(euLabPasses);

            assertEquals(10, testEntries.size());

            assertThatTestEntry(testEntries.get(9)).isNegative().isPcr().hasSampleTime(test1SampleTime);
            assertThatTestEntry(testEntries.get(8)).isNegative().isPcr().hasSampleTime(test2SampleTime);
            assertThatTestEntry(testEntries.get(7)).isNegative().isPcr().hasSampleTime(test4SampleTime);
            assertThatTestEntry(testEntries.get(6)).isNegative().isPcr().hasSampleTime(test5SampleTime);
            assertThatTestEntry(testEntries.get(5)).isNegative().isPcr().hasSampleTime(test6SampleTime);
            assertThatTestEntry(testEntries.get(4)).isNegative().isAG().hasSampleTime(test8SampleTime).hasManufacturer(MANUFACTURER_HSC);
            assertThatTestEntry(testEntries.get(3)).isNegative().isAG().hasSampleTime(test9SampleTime).hasManufacturer(MANUFACTURER_HSC);
            assertThatTestEntry(testEntries.get(2)).isNegative().isAG().hasSampleTime(test11SampleTime).hasManufacturer(MANUFACTURER_HSC);
            assertThatTestEntry(testEntries.get(1)).isNegative().isAG().hasSampleTime(test13SampleTime).hasManufacturer(MANUFACTURER_HSC);
            assertThatTestEntry(testEntries.get(0)).isNegative().isAG().hasSampleTime(test15SampleTime).hasManufacturer(MANUFACTURER_HSC);
        }

        @Test
        @DisplayName("Special Case - Vaccinated person tests positive")
        void vaccinated_person_tests_positive() throws Exception {

            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(Duration.ofDays(28L));

            dataCacheDriver.withUser(user).invalidateLabCache();
            dataCacheDriver.withUser(user).invalidateVacCache();

            ddvDriver.withCitizen(user)
                .clear()
                .addPfizerVaccination(Date.from(effectuationTime));

            kidsDriver.withUser(user).clearKids()
                .addPositivePcr(LocalDateTime.now().minus(Duration.ofHours(24L)));

            final GetCoronaPassResponse coronaPasses = getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPasses)
                .hasNoEuLabPasses()
                .hasNoEuRecPass()
                .hasEuVacPasses(1)
                .hasNoDkPasses();

            final List<VaccinationEntry> vaccinationEntries = getVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));

            assertEquals(1, vaccinationEntries.size());

            assertThatVaccination(vaccinationEntries.get(0))
                .isPfizerVaccination()
                .hasDateOfVaccination(effectuationTime)
                .isDoseOf(1, 2);
        }
    }
    @Nested
    @DisplayName("CBS DGWS Test")
    class DGWSServiceTest {

        @Test
        void dgws_service_normal_flow() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(Duration.ofDays(28L));

            ddvDriver.withCitizen(user)
                .clear()
                .addPfizerVaccination(Date.from(effectuationTime));

            dataCacheDriver.withUser(user).invalidateVacCache();

            final GetCoronaPassResponse response = Wait.waitFor("invalidation", () -> cbsdgwsDriver.getCoronaPasses(user.cpr()), pass -> pass.getFutureResult() != null);

            fetchDataAndWaitForJobStatusNotification(response.getFutureResult().getJobId());

            final GetCoronaPassResponse coronaPasses = cbsdgwsDriver.getCoronaPasses(user.cpr());

            assertThat(coronaPasses)
                .hasEuVacPasses(1);
        }
    }

    @Nested
    @DisplayName("CBS EU Signing Test")
    class EuSigningTests {

        @Test
        void eu_signing_service_normal_flow() {
            final GetSignedEUCovidTestCertificateRequest request = new GetSignedEUCovidTestCertificateRequest();

            final ZonedDateTime sampleCollected = ZonedDateTime.now(ZoneId.of("Europe/Copenhagen"));

            final TestSample payload = new TestSample();
            payload.setGivenName("Brian");
            payload.setFamilyName("London");
            payload.setDateOfBirth(LocalDate.of(1992, Month.DECEMBER, 17));
            payload.setCountryOfTest(CountryOfTest.DK);
            payload.setDiseaseAgentTargeted(DiseaseAgent.COVID_19);
            payload.setTestingCenter("Billund Lufthavn - Falck A/S");
            payload.setManufacturer(MANUFACTURER_HSC);
            payload.setSampleCollected(sampleCollected);
            payload.setTypeOfTest(TestType.PCR);
            payload.setTestResult(TestResult.NEGATIVE);

            request.setPersonIdentifier("xyz-11234");
            request.setPayload(payload);

            final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);

            assertNotNull(response);

            final DigitalGreenCertificate certificate = getEUPayload(response.getCWT());

            final List<TestEntry> testEntries = getTestEntriesOrderedBySampleTime(certificate);

            assertEquals(1, testEntries.size());

            assertThatTestEntry(testEntries.get(0))
                .isNegative()
                .isPcr()
                .hasManufacturer(MANUFACTURER_HSC)
                .hasSampleTime(sampleCollected.toLocalDateTime());

            assertEquals("1.3.0", response.getSchemaVersion());
            assertEquals("https://github.com/ehn-dcc-development/ehn-dcc-schema", response.getDocumentation());
            assertNotNull(response.getCertificateIdentifier());
        }

        @Test
        void eu_signing_service_manufacturer_not_approved() {
            final GetSignedEUCovidTestCertificateRequest request = new GetSignedEUCovidTestCertificateRequest();

            final TestSample payload = new TestSample();
            payload.setGivenName("Brian");
            payload.setFamilyName("London");
            payload.setDateOfBirth(LocalDate.of(1992, Month.DECEMBER, 17));
            payload.setCountryOfTest(CountryOfTest.DK);
            payload.setDiseaseAgentTargeted(DiseaseAgent.COVID_19);
            payload.setTestingCenter("Billund Lufthavn - Falck A/S");
            payload.setManufacturer(MANUFACTURER_NOT_HSC);
            payload.setSampleCollected(ZonedDateTime.now(ZoneId.of("Europe/Copenhagen")));
            payload.setTypeOfTest(TestType.RAT);
            payload.setTestResult(TestResult.NEGATIVE);

            request.setPersonIdentifier("xyz-11234");
            request.setPayload(payload);

            final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

            assertEquals("Test is not HSC Approved", exception.getFault().getFaultString());
        }
    }

    private List<TestEntry> getTestEntriesOrderedBySampleTime(List<PassJWT> euLabPasses) {
        return euLabPasses.stream()
            .map(PassJWT::getJWT)
            .map(CoronaPassITCase.this::getEUPayload)
            .map(Eudgc::getT)
            .flatMap(List::stream)
            .sorted(Comparator.comparing(TestEntry::getSc))
            .collect(Collectors.toList());
    }

    private List<TestEntry> getTestEntriesOrderedBySampleTime(DigitalGreenCertificate certificate) {
        return certificate.getT().stream()
            .sorted(Comparator.comparing(TestEntry::getSc))
            .collect(Collectors.toList());
    }


    private List<VaccinationEntry> getVaccinationEntriesOrderedByDate(List<PassJWT> vaccinationPasses) {
        return vaccinationPasses.stream()
            .map(PassJWT::getJWT)
            .map(CoronaPassITCase.this::getEUPayload)
            .flatMap(cert -> cert.getV().stream())
            .sorted(Comparator.comparing(VaccinationEntry::getDt))
            .collect(Collectors.toList());
    }

    private GetCoronaPassResponse waitForInvalidationAndGetFutureResult(IdwsUser idwsUser) throws InterruptedException {
        return waitForInvalidationAndGetFutureResult(idwsUser, idwsUser);
    }

    private GetCoronaPassResponse waitForInvalidationAndGetFutureResult(IdwsUser caller, IdwsUser requestUser) throws InterruptedException {
        return Wait.waitFor("passport to become invalidated", () -> {
            dataCacheInvalidatorDriver.triggerInvalidation();
            return cbsDriver.getCoronaPass(caller, requestUser.cpr());
        }, r -> r.getFutureResult() != null);
    }

    private void clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser user) throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
        ddvDriver.withCitizen(user).clear()
            .addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS))).deleteVaccination();

        final GetCoronaPassResponse response = waitForInvalidationAndGetFutureResult(user);
        fetchDataAndWaitForJobStatusNotification(response.getFutureResult().getJobId());
    }

    private GetCoronaPassResponse getCoronaPassesAfterInvalidation(IdwsUser idwsUser) throws InterruptedException {
        final String jobId = waitForInvalidationAndGetFutureResult(idwsUser).getFutureResult().getJobId();
        fetchDataAndWaitForJobStatusNotification(jobId);
        return getCoronaPasses(idwsUser);
    }

    private GetCoronaPassResponse getCoronaPasses(IdwsUser idwsUser) throws InterruptedException {
        GetCoronaPassResponse coronaPass = cbsDriver.getCoronaPass(idwsUser);

        if (coronaPass.getFutureResult() != null) {
            fetchDataAndWaitForJobStatusNotification(coronaPass.getFutureResult().getJobId());
            coronaPass = cbsDriver.getCoronaPass(idwsUser);
        }

        return coronaPass;
    }


    private JobStatusNotification fetchDataAndWaitForJobStatusNotification(String jobId) throws InterruptedException {
        return Wait.waitFor("NAS message with jobId " + jobId + " to arrive", () -> {
            dataFetcherDriver.triggerDataFetch();
            final List<JobStatusNotification> notifications = nasDriver.getNotifications();
            notifications.forEach(m -> System.out.println(m.getJobId() + " --> " + m.getJobStatus()));
            return notifications.stream()
                .filter(m -> m.getJobId().equals(jobId))
                .findAny().orElse(null);

        }, Objects::nonNull);
    }


    private void validateTokensAreValidAndFresh(List<PassJWT> passList) {
        for (PassJWT passJWT : passList) {

            if (passJWT.getType().startsWith("eu")) {
                // This will validate the signature:
                DigitalGreenCertificate dgc = getEUPayload(passJWT.getJWT());
                assertNotNull(dgc);
            }

            Cwt cwt = getCwtFromToken(passJWT.getJWT().substring(4));

            assertTrue(
                cwt.getIssuedAt().isAfter(Instant.now().minus(TOKEN_ISSUED_AT_SLACK)),
                "Token of type " + passJWT.getType() + " doesn't seem to be a freshly issued token - was issued at " + cwt.getIssuedAt()
            );
        }
    }

    private Cwt getCwtFromToken(String token) {
        if (token.startsWith(DGCConstants.DGC_V1_HEADER)) {
            token = token.substring(4);
        }

        byte[] compressedCwt = Base45.getDecoder().decode(token);

        byte[] cwtBytes = new byte[0];
        try {
            cwtBytes = Zlib.decompress(compressedCwt, false);
        } catch (ZipException e) {
            throw new RuntimeException("Zip decompression failed", e);
        }
        CoseSign1_Object coseObject = CoseSign1_Object.decode(cwtBytes);

        return coseObject.getCwt();
    }


    private DigitalGreenCertificate getEUPayload(String token) {
        try {
            return decoder.decode(token);
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    private DigitalGreenCertificate getDKINFOPayload(GetCoronaPassResponse response) {
        return getDKINFOPayload(CwtTestUtils.getFirstDKInfoPass(response).getJWT());
    }

    private DigitalGreenCertificate getDKINFOPayload(String jwt) {
        return getEUPayload(jwt.replaceFirst("DK3:", DGCConstants.DGC_V1_HEADER));
    }
}
