/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.util;

import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.KeyValuePair;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;

import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

public class CwtTestUtils {
    private CwtTestUtils() {
    }


    public static List<PassJWT> getDkPasses(GetCoronaPassResponse response) {
        return getDkPasses(response.getCoronaPasses().getPass());
    }

    public static List<PassJWT> getDkPasses(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DK") || pass.getType().equals("DKPLUS"))
            .collect(Collectors.toList());
    }

    public static PassJWT getDKPass(GetCoronaPassResponse response) {
        return getDKPass(response.getCoronaPasses().getPass());
    }

    public static PassJWT getDKPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DK"))
            .findFirst()
            .orElse(null);
    }

    public static PassJWT getDKPlusPass(GetCoronaPassResponse response) {
        return getDKPlusPass(response.getCoronaPasses().getPass());
    }

    public static PassJWT getDKPlusPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DKPLUS"))
            .findFirst()
            .orElse(null);
    }

    public static PassJWT getFirstDKInfoPass(GetCoronaPassResponse response) {
        return getFirstDKInfoPass(response.getCoronaPasses().getPass());
    }

    public static List<PassJWT> getDKInfoPass(GetCoronaPassResponse response) {
        return response.getCoronaPasses().getPass().stream()
            .filter(pass -> pass.getType().equals("DKINFO"))
            .collect(Collectors.toList());
    }

    public static List<PassJWT> getDKInfoPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DKINFO"))
            .collect(Collectors.toList());
    }

    public static PassJWT getFirstDKInfoPass(List<PassJWT> passJWTList) {
        return passJWTList.stream()
            .filter(pass -> pass.getType().equals("DKINFO"))
            .findFirst()
            .orElse(null);
    }


    public static List<PassJWT> getEuVacPasses(GetCoronaPassResponse passport) {
        return getEuVacPasses(passport.getCoronaPasses().getPass());
    }

    public static List<PassJWT> getEuVacPasses(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("vac")))
            .collect(Collectors.toList());
    }

    public static List<PassJWT> getEuLabPasses(GetCoronaPassResponse passport) {
        return getEuLabPasses(passport.getCoronaPasses().getPass());
    }


    public static List<PassJWT> getEuLabPasses(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("lab")))
            .collect(Collectors.toList());
    }

    public static List<PassJWT> getEuRecPasses(GetCoronaPassResponse passport) {
        return getEuRecPasses(passport.getCoronaPasses().getPass());
    }


    public static List<PassJWT> getEuRecPasses(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("rec")))
            .collect(Collectors.toList());
    }

    public static List<PassJWT> getPassesForCustody(GetCoronaPassResponse passport, String custody) {
        return getPassesForCustody(passport.getCoronaPasses().getPass(), custody);
    }

    public static List<PassJWT> getPassesForCustody(List<PassJWT> passList, String custody) {
        return passList.stream()
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("Custody") && prop.getValue().equals(custody)))
            .collect(Collectors.toList());
    }

    public static Set<String> getCustodies(List<PassJWT> passList) {
        return passList.stream()
            .map(CwtTestUtils::getCustody)
            .filter(Objects::nonNull)
            .collect(Collectors.toSet());
    }

    public static String getCustody(PassJWT passJWT) {
        return passJWT.getProperties().stream()
            .filter(p -> p.getKey().equals("Custody"))
            .findAny()
            .map(KeyValuePair::getValue).orElse(null);
    }
}
