/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.webservice;

import com.google.common.truth.Correspondence;
import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.util.CoronaPassAssertions;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.FutureResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassList;
import org.apache.commons.lang3.StringUtils;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.DGCDecoder;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.function.Consumer;

import static org.junit.jupiter.api.Assertions.assertNull;

@SuppressWarnings({"unused", "UnusedReturnValue"})
public class ResponseTester {
    private final List<PassJWT> passList;
    private final FutureResult futureResult;
    private DGCDecoder decoder = null;

    private ResponseTester(GetCoronaPassResponse response, DGCDecoder decoder) {
        this(
            response.getFutureResult(),
            Optional.of(response)
                .map(GetCoronaPassResponse::getCoronaPasses)
                .map(PassList::getPass)
                .orElse(Collections.emptyList())
        );
        this.decoder = decoder;
    }

    private ResponseTester(FutureResult futureResult, List<PassJWT> passList) {
        this.passList = passList;
        this.futureResult = futureResult;
    }

    private ResponseTester(FutureResult futureResult, List<PassJWT> passList, DGCDecoder decoder) {
        this.passList = passList;
        this.futureResult = futureResult;
        this.decoder = decoder;
    }

    public static ResponseTester assertThat(GetCoronaPassResponse response) {
        return new ResponseTester(Objects.requireNonNull(response), null);
    }

    public static ResponseTester assertThat(GetCoronaPassResponse response, DGCDecoder decoder) {
        return new ResponseTester(Objects.requireNonNull(response), Objects.requireNonNull(decoder));
    }


    public ResponseTester forCustody(String custody, Consumer<ResponseTester> tester) {
        tester.accept(new ResponseTester(futureResult, CwtTestUtils.getPassesForCustody(passList, custody), decoder));
        return this;
    }

    public ResponseTester isNotFutureResult() {
        assertNull(futureResult, "Response contained a FutureResult");
        return this;
    }

    public ResponseTester hasDkPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getDkPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasDkInfoWithName(String name) {
        final String given = StringUtils.substringBefore(name, " ");
        final String family = StringUtils.substringAfter(name, " ");
        return hasDkInfoWithName(given, family);
    }

    public ResponseTester hasDkInfoWithName(String given, String family) {
        isNotFutureResult();
        final List<PassJWT> dkInfoPass = CwtTestUtils.getDKInfoPass(passList);
        Truth.assertThat(dkInfoPass).hasSize(1);
        CoronaPassAssertions.assertHasName(given, family, decoder, dkInfoPass);
        return this;
    }

    public ResponseTester hasNoCustodyChildPasses() {
        Truth.assertThat(CwtTestUtils.getCustodies(passList)).containsExactly("Parent");
        return this;
    }

    public ResponseTester hasNoDkPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getDkPasses(passList)).isEmpty();
        return this;
    }

    public ResponseTester hasEuVacPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuVacPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasNoEuVacPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuVacPasses(passList)).isEmpty();
        return this;
    }

    public ResponseTester hasEuLabPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuLabPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasEuLabPassWithSampleTime(LocalDateTime samplingTime) {
        ZoneOffset samplingZoneOffset = ZoneId.systemDefault().getRules().getOffset(samplingTime);
        isNotFutureResult();
        final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(passList);
        Correspondence<PassJWT, LocalDateTime> matchingSamplingTime = Correspondence.from((PassJWT p, LocalDateTime s) -> {
            try {
                final DigitalGreenCertificate digitalGreenCertificate = decoder.decode(p.getJWT());
                return digitalGreenCertificate.getT().stream().anyMatch(t ->
                    t.getSc().truncatedTo(ChronoUnit.SECONDS).equals(samplingTime.toInstant(samplingZoneOffset).truncatedTo(ChronoUnit.SECONDS))
                );
            } catch (Exception e) {
                throw new RuntimeException("decoding failed", e);
            }
        }, "EU PassJWT has samplingTime equal to");

        Truth.assertThat(euLabPasses).comparingElementsUsing(matchingSamplingTime).contains(samplingTime);
        return this;
    }

    public ResponseTester hasNoEuLabPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuLabPasses(passList)).isEmpty();
        return this;
    }


    public ResponseTester hasEuRecPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuRecPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasNoEuRecPass() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuRecPasses(passList)).isEmpty();
        return this;
    }


    public ResponseTester allCwtsWithNameHasName(String name) {
        final ArrayList<PassJWT> passListCopy = new ArrayList<>(passList);
        passListCopy.remove(CwtTestUtils.getDKPass(passListCopy));
        CoronaPassAssertions.assertHasName(name, decoder, passListCopy);
        return this;
    }

}
